from pathlib import Path
import shutil

from mayan.apps.storage.utils import TemporaryDirectory, mkdtemp
from mayan.apps.testing.tests.base import BaseTestCase
from mayan.apps.testing.tests.utils import mute_stdout

from ..classes import JavaScriptDependency
from ..exceptions import DependenciesException

from .mocks import TestDependency
from .literals import TEST_TAR_CVE_2007_4559_FILENAME, TEST_TAR_CVE_2007_4559_PATH


class DependencyClassTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.test_replace_text = 'replaced_text'

        self.temporary_directory = mkdtemp()
        self.path_temporary_directory = Path(self.temporary_directory)
        self.path_test_file = self.path_temporary_directory / 'test_file.css'

        with self.path_test_file.open(mode='w') as file_object:
            file_object.write(
                '@import url("https://fonts.googleapis.com/css?family=Lato:400,700,400italic");'
            )
        self._test_dependency = TestDependency(
            name='test_dependency', module=__name__
        )

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(path=self.temporary_directory, ignore_errors=True)

    def _patch_test_file(self):
        replace_list = [
            {
                'filename_pattern': '*',
                'content_patterns': [
                    {
                        'search': '"https://fonts.googleapis.com/css?family=Lato:400,700,400italic"',
                        'replace': self.test_replace_text,
                    }
                ]
            }
        ]

        with mute_stdout():
            self._test_dependency.patch_files(
                path=self.temporary_directory, replace_list=replace_list
            )

        with self.path_test_file.open(mode='r') as file_object:
            self.final_text = file_object.read()

    def test_file_patching(self):
        self._patch_test_file()

        self.assertEqual(
            self.final_text, '@import url({});'.format(self.test_replace_text)
        )


class DependencyClassCVE_2007_4559TestCase(BaseTestCase):
    def test_path_travesal_detection(self):
        with TemporaryDirectory() as temporary_directory:
            (Path(temporary_directory) / 'package').mkdir()
            with Path(TEST_TAR_CVE_2007_4559_PATH).open(mode='rb') as test_source_file_object:
                with (Path(temporary_directory) / TEST_TAR_CVE_2007_4559_FILENAME).open(mode='wb') as test_destination_file_object:
                    shutil.copyfileobj(fsrc=test_source_file_object, fdst=test_destination_file_object)

            test_dependency = JavaScriptDependency(
                label='Label', module=__name__,
                name='test_repository', version_string='=1.0'
            )
            test_dependency.path_cache = temporary_directory
            test_dependency.version_metadata = {
                'dist': {'tarball': TEST_TAR_CVE_2007_4559_FILENAME}
            }

            with self.assertRaises(expected_exception=DependenciesException):
                test_dependency.extract()



import logging

from django.db import models, transaction
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.common.signals import signal_mayan_pre_save
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter
from mayan.apps.storage.classes import DefinedStorageLazy

from ..events import event_document_file_deleted, event_document_file_edited
from ..literals import STORAGE_NAME_DOCUMENT_FILES
from ..managers import DocumentFileManager, ValidDocumentFileManager

from .document_file_model_mixins import DocumentFileBusinessLogicMixin
from .document_models import Document
from .model_mixins import HooksModelMixin

__all__ = ('DocumentFile', 'DocumentFileSearchResult')
logger = logging.getLogger(name=__name__)


def upload_to(instance, filename):
    return instance.document.document_type.get_upload_filename(
        instance=instance, filename=filename
    )


class DocumentFile(
    DocumentFileBusinessLogicMixin, ExtraDataModelMixin, HooksModelMixin,
    models.Model
):
    """
    Model that describes a document file and its properties
    Fields:
    * mimetype - File mimetype. MIME types are a standard way to describe the
    format of a file, in this case the file format of the document.
    Some examples: "text/plain" or "image/jpeg". Mayan uses this to determine
    how to render a document's file. More information:
    http://www.freeformatter.com/mime-types-list.html
    * encoding - File Encoding. The filesystem encoding of the document's
    file: binary 7-bit, binary 8-bit, text, base64, etc.
    * checksum - A hash/checkdigit/fingerprint generated from the document's
    binary data. Only identical documents will have the same checksum. If a
    document is modified after upload it's checksum will not match, used for
    detecting file tampering among other things.
    """
    _hooks_pre_create = []
    _pre_open_hooks = []
    _pre_save_hooks = []
    _post_save_hooks = []

    document = models.ForeignKey(
        on_delete=models.CASCADE, related_name='files', to=Document,
        verbose_name=_('Document')
    )
    timestamp = models.DateTimeField(
        auto_now_add=True, db_index=True, help_text=_(
            'The server date and time when the document file was processed.'
        ), verbose_name=_('Timestamp')
    )
    comment = models.TextField(
        blank=True, default='', help_text=_(
            'An optional short text describing the document file.'
        ), verbose_name=_('Comment')
    )
    # File related fields.
    file = models.FileField(
        storage=DefinedStorageLazy(name=STORAGE_NAME_DOCUMENT_FILES),
        upload_to=upload_to, verbose_name=_('File')
    )
    filename = models.CharField(
        blank=True, max_length=255, verbose_name=_('Filename')
    )
    mimetype = models.CharField(
        blank=True, editable=False, help_text=_(
            'The document file\'s file mimetype. MIME types are a '
            'standard way to describe the format of a file, in this case '
            'the file format of the document. Some examples: "text/plain" '
            'or "image/jpeg". '
        ), max_length=255, null=True, verbose_name=_('MIME type')
    )
    encoding = models.CharField(
        blank=True, editable=False, help_text=_(
            'The document file file encoding. binary 7-bit, binary 8-bit, '
            'text, base64, etc.'
        ), max_length=64, null=True, verbose_name=_('Encoding')
    )
    checksum = models.CharField(
        blank=True, db_index=True, editable=False, help_text=(
            'A hash/checkdigit/fingerprint generated from the document\'s '
            'binary data. Only identical documents will have the same '
            'checksum.'
        ), max_length=64, null=True, verbose_name=_('Checksum')
    )
    size = models.PositiveBigIntegerField(
        blank=True, db_index=True, editable=False, help_text=(
            'The size of the file in bytes.'
        ), null=True, verbose_name=_('Size')
    )

    class Meta:
        ordering = ('timestamp',)
        verbose_name = _('Document file')
        verbose_name_plural = _('Document files')

    objects = DocumentFileManager()
    valid = ValidDocumentFileManager()

    def __str__(self):
        return self.get_label()

    @method_event(
        action_object='document',
        event_manager_class=EventManagerMethodAfter,
        event=event_document_file_edited,
        target='self'
    )
    def _save(self, *args, **kwargs):
        user = getattr(self, '_event_actor', None)

        try:
            self.execute_pre_save_hooks()

            signal_mayan_pre_save.send(
                instance=self, sender=DocumentFile, user=user
            )

            result = super().save(*args, **kwargs)

            DocumentFile._execute_hooks(
                hook_list=DocumentFile._post_save_hooks,
                instance=self
            )
        except Exception as exception:
            logger.error(
                'Error saving document file for document "%s"; %s',
                self.document, exception, exc_info=True
            )
            raise
        else:
            return result

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_document_file_deleted,
        target='document'
    )
    def delete(self, user=None, *args, **kwargs):
        if user:
            self._event_actor = user

        for page in self.pages.all():
            page._event_actor = getattr(self, '_event_actor', None)
            page.delete()

        name = self.file.name
        self.file.close()
        self.file.storage.delete(name=name)
        self.cache_partition.delete()

        with transaction.atomic():
            result = super().delete(*args, **kwargs)

            self.document.file_latest = self.get_document_file_latest()

            if self.document.files.exclude(pk=self.pk).count() == 0:
                self.document.is_stub = False

            self.document._event_ignore = True
            self.document.save(
                update_fields=('file_latest', 'is_stub')
            )

            return result

    def get_absolute_url(self):
        return reverse(
            viewname='documents:document_file_preview', kwargs={
                'document_file_id': self.pk
            }
        )

    def natural_key(self):
        return (
            self.checksum, self.document.natural_key()
        )
    natural_key.dependencies = ['documents.Document']

    def save(self, skip_introspection=False, *args, **kwargs):
        """
        Overloaded save method that updates the document file's checksum,
        mimetype, and page count when created.
        """
        self._event_keep_attributes = ('_event_actor',)
        new_document_file = not self.pk

        if new_document_file:
            result = self._create(*args, **kwargs)
            if not skip_introspection:
                self._introspect()
            return result
        else:
            return self._save(*args, **kwargs)


class DocumentFileSearchResult(DocumentFile):
    class Meta:
        proxy = True



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Theme',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'label', models.CharField(
                        db_index=True, help_text='A short text describing '
                        'the theme.', max_length=128, unique=True,
                        verbose_name='Label'
                    )
                ),
                (
                    'stylesheet', models.TextField(
                        blank=True, help_text='The CSS stylesheet to '
                        'change the appearance of the different user '
                        'interface elements.', verbose_name='Stylesheet'
                    )
                ),
            ],
            options={
                'verbose_name': 'Theme',
                'verbose_name_plural': 'Themes',
                'ordering': ('label',),
            },
        ),
        migrations.CreateModel(
            name='UserThemeSetting',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'theme', models.ForeignKey(
                        blank=True, null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='user_setting', to='appearance.Theme',
                        verbose_name='Theme'
                    )
                ),
                (
                    'user', models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='theme_settings',
                        to=settings.AUTH_USER_MODEL, verbose_name='User'
                    )
                ),
            ],
            options={
                'verbose_name': 'User theme setting',
                'verbose_name_plural': 'User theme settings',
            },
        ),
    ]



from mayan.apps.acls.models import AccessControlList


class OwnerPlusFilteredQuerysetViewMixin:
    def get_source_queryset(self):
        queryset = super().get_source_queryset()
        queryset_user = queryset.filter(user=self.request.user)

        queryset = queryset_user | AccessControlList.objects.restrict_queryset(
            permission=self.optional_object_permission,
            queryset=queryset, user=self.request.user
        )

        return queryset.distinct()



from django.db import migrations, models
import django.db.models.deletion

import mayan.apps.common.validators


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('converter', '0013_auto_20180823_2353'),
    ]

    operations = [
        migrations.CreateModel(
            name='LayerTransformation',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'order', models.PositiveIntegerField(
                        blank=True, db_index=True, default=0,
                        help_text='Order in which the transformations will '
                        'be executed. If left unchanged, an automatic order '
                        'value will be assigned.', verbose_name='Order'
                    )
                ),
                (
                    'name', models.CharField(
                        choices=[
                            ('crop', 'Crop: left, top, right, bottom'),
                            ('draw_rectangle', 'Draw rectangle: left, top, right, bottom, fillcolor, outlinecolor, outlinewidth'),
                            ('draw_rectangle_percent', 'Draw rectangle (percents coordinates): left, top, right, bottom, fillcolor, outlinecolor, outlinewidth'),
                            ('flip', 'Flip'),
                            ('gaussianblur', 'Gaussian blur: radius'),
                            ('lineart', 'Line art'), ('mirror', 'Mirror'),
                            ('redaction_percent', 'Redaction: left, top, right, bottom'),
                            ('resize', 'Resize: width, height'),
                            ('rotate', 'Rotate: degrees, fillcolor'),
                            ('rotate180', 'Rotate 180 degrees'),
                            ('rotate270', 'Rotate 270 degrees'),
                            ('rotate90', 'Rotate 90 degrees'),
                            ('unsharpmask', 'Unsharp masking: radius, percent, threshold'),
                            ('zoom', 'Zoom: percent')
                        ], max_length=128, verbose_name='Name'
                    )
                ),
                (
                    'arguments', models.TextField(
                        blank=True, help_text='Enter the arguments for '
                        'the transformation as a YAML dictionary. '
                        'ie: {"degrees": 180}', validators=[
                            mayan.apps.common.validators.YAMLValidator()
                        ], verbose_name='Arguments'
                    )
                ),
                (
                    'enabled', models.BooleanField(
                        default=True, verbose_name='Enabled'
                    )
                ),
            ],
            options={
                'ordering': ('object_layer__stored_layer__order', 'order'),
                'verbose_name': 'Layer transformation',
                'verbose_name_plural': 'Layer transformations',
            },
        ),
        migrations.CreateModel(
            name='ObjectLayer',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                ('object_id', models.PositiveIntegerField()),
                (
                    'enabled', models.BooleanField(
                        default=True, verbose_name='Enabled'
                    )
                ),
                (
                    'content_type', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='contenttypes.ContentType'
                    )
                ),
            ],
            options={
                'ordering': ('stored_layer__order',),
                'verbose_name': 'Object layer',
                'verbose_name_plural': 'Object layers',
            },
        ),
        migrations.CreateModel(
            name='StoredLayer',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                (
                    'name', models.CharField(
                        max_length=64, unique=True, verbose_name='Name'
                    )
                ),
                (
                    'order', models.PositiveIntegerField(
                        db_index=True, unique=True, verbose_name='Order'
                    )
                ),
            ],
            options={
                'ordering': ('order',),
                'verbose_name': 'Stored layer',
                'verbose_name_plural': 'Stored layers',
            },
        ),
        migrations.AlterField(
            model_name='transformation',
            name='name',
            field=models.CharField(
                choices=[
                    ('crop', 'Crop: left, top, right, bottom'),
                    ('draw_rectangle', 'Draw rectangle: left, top, right, bottom, fillcolor, outlinecolor, outlinewidth'),
                    ('draw_rectangle_percent', 'Draw rectangle (percents coordinates): left, top, right, bottom, fillcolor, outlinecolor, outlinewidth'),
                    ('flip', 'Flip'),
                    ('gaussianblur', 'Gaussian blur: radius'),
                    ('lineart', 'Line art'), ('mirror', 'Mirror'),
                    ('redaction_percent', 'Redaction: left, top, right, bottom'),
                    ('resize', 'Resize: width, height'),
                    ('rotate', 'Rotate: degrees, fillcolor'),
                    ('rotate180', 'Rotate 180 degrees'),
                    ('rotate270', 'Rotate 270 degrees'),
                    ('rotate90', 'Rotate 90 degrees'),
                    ('unsharpmask', 'Unsharp masking: radius, percent, threshold'),
                    ('zoom', 'Zoom: percent')
                ], max_length=128, verbose_name='Name'
            ),
        ),
        migrations.AddField(
            model_name='objectlayer',
            name='stored_layer',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='object_layers', to='converter.StoredLayer',
                verbose_name='Stored layer'
            ),
        ),
        migrations.AddField(
            model_name='layertransformation',
            name='object_layer',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='transformations', to='converter.ObjectLayer',
                verbose_name='Object layer'
            ),
        ),
        migrations.AlterUniqueTogether(
            name='objectlayer',
            unique_together={
                ('content_type', 'object_id', 'stored_layer')
            },
        ),
        migrations.AlterUniqueTogether(
            name='layertransformation',
            unique_together={('object_layer', 'order')},
        ),
    ]



from django.test import RequestFactory

from ..dashboard_widgets import DashboardWidgetUserRecentlyCreatedDocuments
from ..permissions import permission_document_view

from .base import GenericDocumentViewTestCase


class DashboardWidgetUserRecentlyCreatedDocumentsTestCase(
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_widget_no_permission(self):
        widget = DashboardWidgetUserRecentlyCreatedDocuments()

        request = RequestFactory()
        request.user = self._test_case_user
        widget.request = request

        self.assertTrue(self._test_document not in widget.get_object_list())

    def test_widget_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        widget = DashboardWidgetUserRecentlyCreatedDocuments()

        request = RequestFactory()
        request.user = self._test_case_user
        widget.request = request

        self.assertTrue(self._test_document in widget.get_object_list())



from django.utils.translation import ugettext_lazy as _

from mayan.apps.dependencies.classes import BinaryDependency
from mayan.apps.sources.settings import setting_backend_arguments

from .literals import DEFAULT_BINARY_SCANIMAGE_PATH

BinaryDependency(
    label='SANE scanimage', help_text=_(
        'Utility provided by the SANE package. Used to control the scanner '
        'and obtained the scanned document image.'
    ), module=__name__, name='scanimage',
    path=setting_backend_arguments.value.get(
        'mayan.apps.source_sane_scanners.source_backends.SourceBackendSANEScanner', {}
    ).get('scanimage_path', DEFAULT_BINARY_SCANIMAGE_PATH)
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Common'), name='common'
)

permission_object_copy = namespace.add_permission(
    label=_('Copy object'), name='object_copy'
)



from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404

from mayan.apps.permissions.models import StoredPermission
from mayan.apps.permissions.serializers import PermissionSerializer
from mayan.apps.rest_api.api_view_mixins import ExternalContentTypeObjectAPIViewMixin
from mayan.apps.rest_api import generics

from .classes import ModelPermission
from .permissions import permission_acl_edit, permission_acl_view
from .serializers import (
    ACLPermissionAddSerializer, ACLPermissionRemoveSerializer, ACLSerializer
)


class APIACLListView(
    ExternalContentTypeObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the object's access control lists
    post: Create a new access control list for the selected object.
    """
    mayan_external_object_permissions = {
        'GET': (permission_acl_view,),
        'POST': (permission_acl_edit,)
    }
    ordering_fields = ('id', 'role')
    serializer_class = ACLSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'content_object': self.get_external_object()
        }

    def get_source_queryset(self):
        return self.get_external_object().acls.all()


class APIACLDetailView(
    ExternalContentTypeObjectAPIViewMixin, generics.RetrieveDestroyAPIView
):
    """
    delete: Delete the selected access control list.
    get: Returns the details of the selected access control list.
    """
    lookup_url_kwarg = 'acl_id'
    mayan_external_object_permissions = {
        'DELETE': (permission_acl_edit,),
        'GET': (permission_acl_view,)
    }
    serializer_class = ACLSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_source_queryset(self):
        return self.get_external_object().acls.all()


class APIACLPermissionAddView(
    ExternalContentTypeObjectAPIViewMixin, generics.ObjectActionAPIView
):
    """
    post: Add a permission to an ACL.
    """
    lookup_url_kwarg = 'acl_id'
    mayan_external_object_permissions = {
        'POST': (permission_acl_edit,)
    }
    serializer_class = ACLPermissionAddSerializer

    def get_source_queryset(self):
        return self.get_external_object().acls

    def object_action(self, obj, request, serializer):
        obj.permissions_add(
            queryset=StoredPermission.objects.filter(
                pk=serializer.validated_data['permission'].stored_permission.pk
            ), user=self.request.user
        )


class APIACLPermissionListView(
    ExternalContentTypeObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns the access control list permission list.
    post: Add a new permission to the selected access control list.
    """
    mayan_external_object_permissions = {
        'GET': (permission_acl_view,),
        'POST': (permission_acl_edit,)
    }
    serializer_class = PermissionSerializer

    def get_acl(self):
        return get_object_or_404(
            klass=self.get_external_object().acls, pk=self.kwargs['acl_id']
        )

    def get_source_queryset(self):
        return self.get_acl().permissions.all()


class APIACLPermissionRemoveView(
    ExternalContentTypeObjectAPIViewMixin, generics.ObjectActionAPIView
):
    """
    post: Remove a permission from an ACL.
    """
    lookup_url_kwarg = 'acl_id'
    mayan_external_object_permissions = {
        'POST': (permission_acl_edit,)
    }
    serializer_class = ACLPermissionRemoveSerializer

    def get_source_queryset(self):
        return self.get_external_object().acls

    def object_action(self, obj, request, serializer):
        obj.permissions_remove(
            queryset=StoredPermission.objects.filter(
                pk=serializer.validated_data['permission'].stored_permission.pk
            ), user=self.request.user
        )


class APIClassPermissionList(generics.ListAPIView):
    """
    Returns a list of all the available permissions for a class.
    """
    serializer_class = PermissionSerializer

    def get_content_type(self):
        return get_object_or_404(
            klass=ContentType, app_label=self.kwargs['app_label'],
            model=self.kwargs['model_name']
        )

    def get_source_queryset(self):
        return ModelPermission.get_for_class(
            klass=self.get_content_type().model_class()
        )



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('announcements', '0002_auto_20210212_1022'),
        ('motd', '0005_auto_20160510_0025'),
    ]

    operations = [
        migrations.DeleteModel(
            name='Message',
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('quotas', '0001_initial')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='quota',
            options={
                'ordering': ('id',), 'verbose_name': 'Quota',
                'verbose_name_plural': 'Quotas'
            },
        )
    ]



from rest_framework.permissions import IsAuthenticated

from mayan.apps.rest_api import generics

from .classes import AJAXTemplate
from .serializers import AJAXTemplateSerializer


class APITemplateDetailView(generics.RetrieveAPIView):
    """
    Returns the selected partial template details.
    get: Retrieve the details of the partial template.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = AJAXTemplateSerializer

    def get_object(self):
        return AJAXTemplate.get(
            name=self.kwargs['name']
        ).render(
            request=self.request
        )


class APITemplateListView(generics.ListAPIView):
    """
    Returns a list of all the available templates.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = AJAXTemplateSerializer

    def get_source_queryset(self):
        return AJAXTemplate.all(
            rendered=True, request=self.request
        )



from django.conf.urls import url

from .views import UserLocaleProfileDetailView, UserLocaleProfileEditView

urlpatterns = [
    url(
        regex=r'^user/(?P<user_id>\d+)/locale/$',
        name='user_locale_profile_detail',
        view=UserLocaleProfileDetailView.as_view()
    ),
    url(
        regex=r'^user/(?P<user_id>\d+)/locale/edit/$',
        name='user_locale_profile_edit',
        view=UserLocaleProfileEditView.as_view()
    )
]



import os

from django.conf import settings

DEFAULT_SIGNATURES_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_SIGNATURES_STORAGE_BACKEND_ARGUMENTS = {
    'location': os.path.join(settings.MEDIA_ROOT, 'document_signatures')
}
RETRY_DELAY = 10
STORAGE_NAME_DOCUMENT_SIGNATURES_DETACHED_SIGNATURE = 'document_signatures__detachedsignature'



import logging

from django.apps import apps
from django.core.exceptions import PermissionDenied
from django.template import Variable, VariableDoesNotExist
from django.urls import Resolver404, resolve

from mayan.apps.permissions import Permission

logger = logging.getLogger(name=__name__)


def factory_condition_queryset_access(
    app_label, model_name, object_permission, callback=None,
    view_permission=None
):
    """
    Return a function that first checks to see if the user has the view
    permission. If not, then filters the objects with the object permission
    and return True if there is at least one item in the filtered queryset.
    This is used to avoid showing a link that ends up in a view with an
    empty results set because the user doesn't have access to any of the
    objects in the queryset.
    """
    def function_condition(context, resolved_object):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )

        Model = apps.get_model(app_label=app_label, model_name=model_name)

        try:
            request = context.request
        except AttributeError:
            # Simple request extraction failed. Might not be a view context.
            # Try alternate method.
            try:
                request = Variable(var='request').resolve(context=context)
            except VariableDoesNotExist:
                # There is no request variable, most probable a 500 in a test
                # view. Don't return anything.
                logger.warning(
                    'No request variable, aborting cascade condition.'
                )
                return ()

        user = request.user

        if view_permission:
            if user.is_authenticated:
                try:
                    Permission.check_user_permissions(
                        permissions=(view_permission,), user=user
                    )
                except PermissionDenied:
                    """
                    Don't raise an error, just ignore and let
                    .restrict_queryset() perform a fine grained filtering.
                    """
                else:
                    if callback:
                        return callback(
                            context=context, resolved_object=resolved_object
                        )
                    else:
                        return True
            else:
                return False

        if user.is_authenticated:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=object_permission, queryset=Model.objects.all(),
                user=user
            )
            if callback:
                return queryset.exists() and callback(
                    context=context, resolved_object=resolved_object
                )
            else:
                return queryset.exists()
        elif callback:
            return False

    return function_condition


def get_content_type_kwargs_factory(
    result_map=None, variable_name='resolved_object'
):
    if not result_map:
        result_map = {
            'app_label': 'app_label',
            'model_name': 'model_name',
            'object_id': 'object_id'
        }

    def get_kwargs(context):
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )

        content_type = ContentType.objects.get_for_model(
            model=context[variable_name]
        )
        return {
            result_map['app_label']: '"{}"'.format(content_type.app_label),
            result_map['model_name']: '"{}"'.format(content_type.model),
            result_map['object_id']: '{}.pk'.format(variable_name)
        }

    return get_kwargs


def get_current_view_name(request):
    current_path = request.META['PATH_INFO']

    # Get sources: view name, view objects.
    try:
        current_view_name = resolve(path=current_path).view_name
    except Resolver404:
        # Can't figure out which view corresponds to this URL.
        # Most likely it is an invalid URL.
        logger.warning(
            'Can\'t figure out which view corresponds to this '
            'URL: %s; aborting menu resolution.', current_path
        )
    else:
        return current_view_name



from mayan.apps.sources.classes import DocumentCreateWizardStep


class SourceBackendMixinInteractive:
    def callback_post_document_create(self, **kwargs):
        super().callback_post_document_create(**kwargs)

        document = kwargs['document']
        query_string = kwargs['query_string']
        source_id = kwargs['source_id']
        user_id = kwargs.get('user_id')

        DocumentCreateWizardStep.post_upload_process(
            document=document, query_string=query_string,
            source_id=source_id, user_id=user_id
        )



from mayan.apps.appearance.classes import Icon

icon_result_list = Icon(driver_name='fontawesome', symbol='search')
icon_search = Icon(driver_name='fontawesome', symbol='search')
icon_search_advanced = Icon(driver_name='fontawesome', symbol='search-plus')
icon_search_again = Icon(driver_name='fontawesome', symbol='sync')
icon_search_backend_reindex = Icon(
    driver_name='fontawesome-layers', data=[
        {
            'class': 'fas fa-circle',
            'transform': 'down-3 right-10',
            'mask': 'fas fa-search'
        },
        {'class': 'far fa-circle', 'transform': 'down-3 right-10'},
        {'class': 'fas fa-search', 'transform': 'flip-h left-3'},
        {'class': 'fas fa-hammer', 'transform': 'shrink-4 down-3 right-10'}
    ]
)
icon_search_form_clear = Icon(
    driver_name='fontawesome', symbol='times-circle'
)
icon_search_submit = Icon(driver_name='fontawesome', symbol='search')



from mayan.apps.documents.events import (
    event_document_created, event_document_type_changed
)
from mayan.apps.documents.permissions import permission_document_change_type
from mayan.apps.documents.tests.base import GenericDocumentTestCase

from ..events import event_workflow_instance_created

from .mixins.workflow_instance_mixins import WorkflowInstanceTestMixin


class WorkflowInstanceModelTestCase(
    WorkflowInstanceTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()

    def test_workflow_launch_on_document_type_change(self):
        self._create_test_document_type()

        self._create_test_document_stub()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_change_type
        )
        self.grant_access(
            obj=self._test_document_types[0],
            permission=permission_document_change_type
        )

        self._clear_events()

        self.assertEqual(
            self._test_document.workflows.count(), 0
        )

        self._test_document.document_type_change(
            document_type=self._test_document_types[0],
            user=self._test_case_user
        )

        self.assertEqual(
            self._test_document.workflows.count(), 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(
            events[0].actor, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(events[0].verb, event_workflow_instance_created.id)

        self.assertEqual(
            events[1].action_object, self._test_document_types[0]
        )
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, self._test_document)
        self.assertEqual(events[1].verb, event_document_type_changed.id)

    def test_workflow_instance_method_get_absolute_url(self):
        self._create_test_document_stub()

        self._test_workflow_instance = self._test_document.workflows.first()

        self._test_workflow_instance.get_absolute_url()

    def test_workflow_template_auto_launch(self):
        self._test_workflow_template.auto_launch = True
        self._test_workflow_template.save()

        self._clear_events()

        self._create_test_document_stub()

        self.assertEqual(
            self._test_document.workflows.count(), 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(
            events[0].actor, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(events[0].verb, event_workflow_instance_created.id)

        self.assertEqual(
            events[1].action_object, self._test_document.document_type
        )
        self.assertEqual(events[1].actor, self._test_document)
        self.assertEqual(events[1].target, self._test_document)
        self.assertEqual(events[1].verb, event_document_created.id)

    def test_workflow_template_no_auto_launch(self):
        self._test_workflow_template.auto_launch = False
        self._test_workflow_template.save()

        self._clear_events()

        self._create_test_document_stub()

        self.assertEqual(
            self._test_document.workflows.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self._test_document.document_type
        )
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

    def test_workflow_template_transition_no_condition(self):
        self._clear_events()

        self._create_test_document_stub()

        self._test_workflow_instance = self._test_document.workflows.first()
        self.assertEqual(
            self._test_workflow_instance.get_transition_choices().count(), 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 2)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(
            events[0].actor, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(events[0].verb, event_workflow_instance_created.id)

        self.assertEqual(
            events[1].action_object, self._test_document.document_type
        )
        self.assertEqual(events[1].actor, self._test_document)
        self.assertEqual(events[1].target, self._test_document)
        self.assertEqual(events[1].verb, event_document_created.id)

    def test_workflow_template_transition_false_condition(self):
        self._create_test_document_stub()

        self._test_workflow_instance = self._test_document.workflows.first()

        self._test_workflow_template_transition.condition = '{{ invalid_variable }}'
        self._test_workflow_template_transition.save()

        self.assertEqual(
            self._test_workflow_instance.get_transition_choices().count(), 0
        )

    def test_workflow_template_transition_true_condition(self):
        self._clear_events()

        self._create_test_document_stub()

        self._test_workflow_instance = self._test_document.workflows.first()

        self._test_workflow_template_transition.condition = '{{ workflow_instance }}'
        self._test_workflow_template_transition.save()

        self.assertEqual(
            self._test_workflow_instance.get_transition_choices().count(), 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(
            events[0].actor, self._test_document.workflows.first()
        )
        self.assertEqual(
            events[0].target, self._test_document.workflows.first()
        )
        self.assertEqual(events[0].verb, event_workflow_instance_created.id)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Redactions'), name='redactions'
)

permission_redaction_create = namespace.add_permission(
    label=_('Create new redactions'), name='redaction_create'
)
permission_redaction_delete = namespace.add_permission(
    label=_('Delete redactions'), name='redaction_delete'
)
permission_redaction_edit = namespace.add_permission(
    label=_('Edit redactions'), name='redaction_edit'
)
permission_redaction_exclude = namespace.add_permission(
    label=_('Exclude redactions'), name='redaction_exclude'
)
permission_redaction_view = namespace.add_permission(
    label=_('View existing redactions'), name='redaction_view'
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('mailer', '0008_migrate_to_credentials')
    ]

    operations = [
        migrations.AlterField(
            field=models.TextField(
                blank=True, help_text='JSON encoded data for the backend '
                'class.', verbose_name='Backend data'
            ), model_name='usermailer', name='backend_data'
        )
    ]



from furl import furl
from PIL import Image

from django.apps import apps
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.locales.utils import to_language

from .events import event_document_version_exported
from .literals import (
    DOCUMENT_VERSION_EXPORT_MESSAGE_BODY,
    DOCUMENT_VERSION_EXPORT_MESSAGE_SUBJECT
)


class DocumentVersionExporter:
    def __init__(self, document_version):
        self.document_version = document_version

    def page_export(self, file_object, page, append=False, resolution=None):
        if not resolution:
            resolution = 300.0

        cache_filename = page.generate_image()
        with page.cache_partition.get_file(filename=cache_filename).open() as image_file_object:
            Image.open(fp=image_file_object).save(
                append=append, format='PDF', fp=file_object,
                resolution=resolution
            )

    def export(self, file_object):
        pages_queryset = self.document_version.pages

        if pages_queryset.exists():
            # Only export the version if there is at least one page.
            export_file_created = False
            for page in pages_queryset:
                if page.content_object:
                    # Ensure only pages that point to actual content are
                    # exported.
                    if not export_file_created:
                        self.page_export(file_object=file_object, page=page)
                    else:
                        self.page_export(append=True, file_object=file_object, page=page)

    def export_to_download_file(
        self, organization_installation_url='', user=None
    ):
        DownloadFile = apps.get_model(
            app_label='storage', model_name='DownloadFile'
        )
        Message = apps.get_model(
            app_label='messaging', model_name='Message'
        )

        download_file = DownloadFile(
            filename='{}.pdf'.format(self.document_version),
            label=_('Document version export to PDF'),
            user=user
        )
        download_file._event_actor = user
        download_file.save()

        with download_file.open(mode='wb+') as file_object:
            self.export(file_object=file_object)

        event_document_version_exported.commit(
            action_object=download_file, actor=user,
            target=self.document_version
        )

        if user:
            download_list_url = furl(organization_installation_url).join(
                reverse(
                    viewname='storage:download_file_list'
                )
            ).tostr()

            download_url = furl(organization_installation_url).join(
                reverse(
                    viewname='storage:download_file_download',
                    kwargs={'download_file_id': download_file.pk}
                )
            ).tostr()

            Message.objects.create(
                sender_object=download_file,
                user=user,
                subject=to_language(
                    language=user.locale_profile.language,
                    promise=DOCUMENT_VERSION_EXPORT_MESSAGE_SUBJECT
                ),
                body=to_language(
                    language=user.locale_profile.language,
                    promise=DOCUMENT_VERSION_EXPORT_MESSAGE_BODY
                ) % {
                    'download_list_url': download_list_url,
                    'download_url': download_url,
                    'document_version': self.document_version,
                }
            )



# -*- coding: utf-8 -*-

TEST_DOCUMENT_VERSION_OCR_CONTENT = 'Mayan EDMS Documentation'
TEST_DOCUMENT_VERSION_OCR_CONTENT_DEU_1 = 'Repository für elektronische Dokumente.'
TEST_DOCUMENT_VERSION_OCR_CONTENT_DEU_2 = 'Es bietet einen'

TEST_DOCUMENT_VERSION_OCR_INDEX_NODE_TEMPLATE = '{{ document.ocr_content|join:" " }}'

TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT = 'test content'
TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED = 'updated content'

TEST_UPDATE_DOCUMENT_PAGE_OCR_ACTION_DOTTED_PATH = 'mayan.apps.ocr.workflow_actions.UpdateDocumentPageOCRAction'



from mayan.apps.smart_settings.classes import Setting
from mayan.apps.smart_settings.tests.mixins import SmartSettingTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..settings import setting_ocr_backend, setting_ocr_backend_arguments


class OCRSettingMigrationTestCase(SmartSettingTestMixin, BaseTestCase):
    def test_ocr_backend_arguments_0001(self):
        test_value = {'location': 'test value'}
        self.test_setting = setting_ocr_backend_arguments
        self.test_config_value = '{}'.format(
            Setting.serialize_value(value=test_value)
        )
        self._create_test_config_file()

        self.assertEqual(
            setting_ocr_backend_arguments.value, test_value
        )

    def test_ocr_backend_arguments_0001_with_dict(self):
        test_value = {'location': 'test value'}
        self.test_setting = setting_ocr_backend_arguments
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_ocr_backend_arguments.value, test_value
        )

    def test_ocr_backend_0002(self):
        test_value = 'test value'
        self.test_setting = setting_ocr_backend
        self.test_config_value = test_value
        self._create_test_config_file()

        self.assertEqual(
            setting_ocr_backend.value,
            'mayan.apps.ocr.backends.tesseract.Tesseract'
        )



from django.apps import apps
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.models.document_models import Document
from mayan.apps.templating.classes import Template

from .events import event_smart_link_edited
from .literals import INCLUSION_AND, INCLUSION_OR


class ResolvedSmartLinkBusinessLogicMixin:
    def get_label_for(self, document):
        return self.get_dynamic_label(document=document) or self.label


class SmartLinkBusinessLogicMixin:
    def document_types_add(self, queryset, user):
        for obj in queryset:
            self.document_types.add(obj)
            event_smart_link_edited.commit(
                action_object=obj, actor=user, target=self
            )

    def document_types_remove(self, queryset, user):
        for obj in queryset:
            self.document_types.remove(obj)
            event_smart_link_edited.commit(
                action_object=obj, actor=user, target=self
            )

    def get_dynamic_label(self, document):
        """
        If the smart links was created using a template label instead of a
        static label, resolve the template and return the result.
        """
        if self.dynamic_label:
            try:
                template = Template(template_string=self.dynamic_label)
                return template.render(context={'document': document})
            except Exception as exception:
                return _(
                    'Error generating dynamic label; %s' % str(
                        exception
                    )
                )
        else:
            return None

    def get_linked_documents_for(self, document):
        """
        Execute the corresponding smart links conditions for the document
        provided and return the resulting document queryset.
        """
        if document.document_type.pk not in self.document_types.values_list('pk', flat=True):
            raise Exception(
                _(
                    'This smart link is not allowed for the selected '
                    'document\'s type.'
                )
            )

        smart_link_query = Q()

        for condition in self.conditions.filter(enabled=True):
            template = Template(template_string=condition.expression)

            condition_query = Q(**{
                '{}__{}'.format(
                    condition.foreign_document_data, condition.operator
                ): template.render(
                    context={'document': document}
                )
            })
            if condition.negated:
                condition_query = ~condition_query

            if condition.inclusion == INCLUSION_AND:
                smart_link_query &= condition_query
            elif condition.inclusion == INCLUSION_OR:
                smart_link_query |= condition_query

        if smart_link_query:
            queryset = Document.objects.filter(smart_link_query)
        else:
            queryset = Document.objects.none()

        return Document.valid.filter(
            pk__in=queryset.values('pk')
        )

    def resolve_for(self, document):
        ResolvedSmartLink = apps.get_model(
            app_label='linking', model_name='ResolvedSmartLink'
        )

        return ResolvedSmartLink(
            smart_link=self, queryset=self.get_linked_documents_for(
                document=document
            )
        )


class SmartLinkConditionBusinessLogicMixin:
    def get_full_label(self):
        return '{} foreign {} {} {} {}'.format(
            self.get_inclusion_display(),
            self.foreign_document_data, _('not') if self.negated else '',
            self.get_operator_display(), self.expression
        )

    get_full_label.short_description = _('Full label')



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        (
            'documents',
            '0084_alter_documenttype_document_stub_pruning_enabled'
        )
    ]

    operations = [
        migrations.AddField(
            model_name='document', name='file_latest',
            field=models.OneToOneField(
                blank=True, null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='document_latest', to='documents.documentfile',
                verbose_name='Latest document file'
            )
        )
    ]



from django.apps import apps

from mayan.apps.documents.models.document_models import Document
from mayan.apps.events.classes import ModelEventType

from ..events import event_index_template_edited


class IndexTemplateBusinessLogicMixin:
    def do_event_triggers_populate(self):
        IndexTemplateEventTrigger = apps.get_model(
            app_label='document_indexing',
            model_name='IndexTemplateEventTrigger'
        )

        entries = []

        for event_type in ModelEventType.get_for_class(klass=Document):
            entries.append(
                IndexTemplateEventTrigger(
                    index_template=self,
                    stored_event_type=event_type.get_stored_event_type()
                )
            )

        IndexTemplateEventTrigger.objects.bulk_create(objs=entries)

    def document_types_add(self, queryset, user):
        for document_type in queryset:
            self.document_types.add(document_type)

            event_index_template_edited.commit(
                action_object=document_type,
                actor=user, target=self
            )

    def document_types_remove(self, queryset, user):
        for document_type in queryset:
            self.document_types.remove(document_type)

            event_index_template_edited.commit(
                action_object=document_type,
                actor=user, target=self
            )

    def delete_index_instance_nodes(self):
        IndexInstanceNode = apps.get_model(
            app_label='document_indexing', model_name='IndexInstanceNode'
        )

        try:
            IndexInstanceNode.objects.filter(
                index_template_node__index=self
            ).delete()
        except IndexInstanceNode.DoesNotExist:
            """Empty index, ignore this exception."""

    def get_document_types_names(self):
        return ', '.join(
            [
                str(document_type) for document_type in self.document_types.all()
            ] or ['None']
        )

    @property
    def index_template_root_node(self):
        """
        Return the root node for this index.
        """
        return self.index_template_nodes.get(parent=None)

    def rebuild(self):
        """
        Delete and reconstruct the index by deleting of all its instance nodes
        and recreating them for the documents whose types are associated with
        this index
        """
        IndexInstance = apps.get_model(
            app_label='document_indexing', model_name='IndexInstance'
        )

        if self.enabled:
            self.delete_index_instance_nodes()

            # Create the new root index instance node.
            self.index_template_root_node.index_instance_nodes.create()

            index_instance = IndexInstance.objects.get(pk=self.pk)
            index_instance.index_instance_root_node
            # Re-index each document with a type associated with this index.
            for document in Document.objects.filter(document_type__in=self.document_types.all()):
                # Evaluate each index template node for each document
                # associated with this index.
                index_instance.document_add(document=document)

    def reset(self):
        self.delete_index_instance_nodes()

        # Create the new root index instance node.
        self.index_template_root_node.initialize_index_instance_root_node()


class IndexTemplateNodeBusinessLogicMixin:
    def get_index_instance_root_node(self):
        return self.index_instance_nodes.get(parent=None)

    def initialize_index_instance_root_node(self):
        self.index_instance_nodes.get_or_create(parent=None)



from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.document_indexing.models.index_instance_models import IndexInstanceNode
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin

from .literals import (
    TEST_WORKFLOW_INDEX_TEMPLATE_EXPRESSION,
    TEST_WORKFLOW_TEMPLATE_LABEL_EDITED,
    TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_INDEX_TEMPLATE_EXPRESSION,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED
)
from .mixins.workflow_instance_mixins import WorkflowInstanceTestMixin
from .mixins.workflow_template_transition_mixins import WorkflowTemplateTransitionTestMixin


class WorkflowInstanceIndexingTestCase(
    IndexTemplateTestMixin, WorkflowInstanceTestMixin,
    GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_WORKFLOW_INDEX_TEMPLATE_EXPRESSION
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)

    def _create_test_workflow_states_and_transitions(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()

    def test_workflow_indexing_workflow_instance_no_initial_state(self):
        self._create_test_document_stub()

        test_workflow_instance = self._test_document.workflows.first()

        value = '{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state()
        )

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )

    def test_workflow_indexing_workflow_instance_initial_state(self):
        self._create_test_workflow_states_and_transitions()
        self._create_test_document_stub()

        test_workflow_instance = self._test_document.workflows.first()

        value = '{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state()
        )

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )

    def test_workflow_indexing_workflow_instance_transition(self):
        self._create_test_workflow_states_and_transitions()
        self._create_test_document_stub()

        test_workflow_instance = self._test_document.workflows.first()

        value = '{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state()
        )

        self._test_document.workflows.first().do_transition(
            transition=self._test_workflow_template_transition
        )

        value_transitioned = '{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state()
        )

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )
        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value_transitioned
            ).exists()
        )


class WorkflowTemplateIndexingTestCase(
    IndexTemplateTestMixin, WorkflowTemplateTransitionTestMixin,
    GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_WORKFLOW_INDEX_TEMPLATE_EXPRESSION
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)

    def _create_test_workflow_states_and_transitions(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()

    def test_workflow_indexing_workflow_template_delete(self):
        self._create_test_workflow_states_and_transitions()
        self._create_test_document_stub()

        self._test_workflow_template.delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
            ).exists()
        )

    def test_workflow_indexing_workflow_template_edit(self):
        self._create_test_workflow_states_and_transitions()
        self._create_test_document_stub()

        test_workflow_instance = self._test_document.workflows.first()

        value = '{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state()
        )

        self._test_workflow_template.label = TEST_WORKFLOW_TEMPLATE_LABEL_EDITED
        self._test_workflow_template.save()

        value_edited = '{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state()
        )

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )
        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value_edited
            ).exists()
        )

    def test_workflow_indexing_workflow_template_state_delete(self):
        self._create_test_workflow_states_and_transitions()
        self._create_test_document_stub()

        self._test_workflow_template_states[0].delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
            ).exists()
        )

    def test_workflow_indexing_workflow_template_state_edit(self):
        self._create_test_workflow_states_and_transitions()
        self._create_test_document_stub()

        test_workflow_instance = self._test_document.workflows.first()

        value = '{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state()
        )

        self._test_workflow_template_states[0].label = TEST_WORKFLOW_TEMPLATE_STATE_LABEL_EDITED
        self._test_workflow_template_states[0].save()

        value_edited = '{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state()
        )

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )
        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value_edited
            ).exists()
        )


class WorkflowTemplateTranstitionIndexingTestCase(
    IndexTemplateTestMixin, WorkflowTemplateTransitionTestMixin,
    GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_WORKFLOW_TEMPLATE_TRANSITION_INDEX_TEMPLATE_EXPRESSION
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)

    def _create_test_workflow_states_and_transitions(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()

    def test_workflow_indexing_workflow_template_transition_delete(self):
        self._create_test_workflow_states_and_transitions()
        self._create_test_document_stub()

        self._test_workflow_template_transition.delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
            ).exists()
        )

    def test_workflow_indexing_workflow_template_transition_edit(self):
        self._create_test_workflow_states_and_transitions()
        self._create_test_document_stub()

        test_workflow_instance = self._test_document.workflows.first()

        value = '{}-{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state(),
            self._test_workflow_template.transitions.first()
        )

        self._test_workflow_template_transition.label = TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL_EDITED
        self._test_workflow_template_transition.save()

        value_edited = '{}-{}-{}'.format(
            self._test_workflow_template.label,
            test_workflow_instance.get_current_state(),
            self._test_workflow_template.transitions.first()
        )

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value
            ).exists()
        )
        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
                value=value_edited
            ).exists()
        )



from django.apps import apps

from .literals import STORAGE_NAME_WORKFLOW_CACHE


def callback_update_workflow_image_cache_size(setting):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')

    cache = Cache.objects.get(
        defined_storage_name=STORAGE_NAME_WORKFLOW_CACHE
    )
    cache.maximum_size = setting.value
    cache.save()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Logging'), name='logging'
)

permission_error_log_entry_delete = namespace.add_permission(
    label=_('Delete error log'), name='error_log_delete'
)
permission_error_log_entry_view = namespace.add_permission(
    label=_('View error log'), name='error_log_view'
)



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..models import Notification
from ..permissions import permission_events_view

from .mixins import (
    EventObjectTestMixin, EventTypeTestMixin, NotificationViewTestMixin
)


class NotificationViewTestCase(
    EventObjectTestMixin, EventTypeTestMixin, NotificationViewTestMixin,
    GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_test_object_with_event_type_and_permission()
        self._create_test_event(target=self._test_object)
        self._create_test_notification()

    def test_notification_list_view_no_permission(self):
        response = self._request_test_notification_list_view()
        self.assertNotContains(
            response=response,
            text=self._test_notification.get_event_type().label,
            status_code=200
        )

    def test_notification_list_view_with_access(self):
        self.grant_access(
            obj=self._test_object, permission=permission_events_view
        )
        response = self._request_test_notification_list_view()
        self.assertContains(
            response=response,
            text=self._test_notification.get_event_type().label,
            status_code=200
        )

    def test_notification_mark_read_all_view_no_permission(self):
        notification_count = Notification.objects.get_unread().count()

        response = self._request_test_notification_mark_read_all_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Notification.objects.get_unread().count(),
            notification_count
        )

    def test_notification_mark_read_all_view_with_access(self):
        self.grant_access(
            obj=self._test_object, permission=permission_events_view
        )

        notification_count = Notification.objects.get_unread().count()

        response = self._request_test_notification_mark_read_all_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Notification.objects.get_unread().count(),
            notification_count - 1
        )

    def test_notification_mark_read_view_no_permission(self):
        notification_count = Notification.objects.get_unread().count()

        response = self._request_test_notification_mark_read()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Notification.objects.get_unread().count(),
            notification_count
        )

    def test_notification_mark_read_view_with_access(self):
        self.grant_access(
            obj=self._test_object, permission=permission_events_view
        )
        notification_count = Notification.objects.get_unread().count()

        response = self._request_test_notification_mark_read()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Notification.objects.get_unread().count(),
            notification_count - 1
        )



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_workflow_template_edited
from ..permissions import permission_workflow_template_edit

from .literals import TEST_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH
from .mixins.workflow_template_state_action_mixins import WorkflowTemplateStateActionViewTestMixin


class WorkflowStateActionViewTestCase(
    WorkflowTemplateStateActionViewTestMixin, GenericViewTestCase
):
    def test_workflow_state_action_create_get_view_no_permission(self):
        action_count = self._test_workflow_template_state.actions.count()

        self._clear_events()

        response = self._request_test_workflow_template_state_action_create_get_view(
            backend_path=TEST_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH
        )
        self.assertEqual(response.status_code, 404)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_action_create_get_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )
        action_count = self._test_workflow_template_state.actions.count()

        self._clear_events()

        response = self._request_test_workflow_template_state_action_create_get_view(
            backend_path=TEST_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH
        )
        self.assertEqual(response.status_code, 200)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_action_create_post_view_no_permission(self):
        action_count = self._test_workflow_template_state.actions.count()

        self._clear_events()

        response = self._request_test_workflow_template_state_action_create_post_view(
            backend_path=TEST_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH
        )
        self.assertEqual(response.status_code, 404)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_action_create_post_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )
        action_count = self._test_workflow_template_state.actions.count()

        self._clear_events()

        response = self._request_test_workflow_template_state_action_create_post_view(
            backend_path=TEST_WORKFLOW_TEMPLATE_STATE_ACTION_DOTTED_PATH
        )
        self.assertEqual(response.status_code, 302)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(),
            action_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object,
            self._test_workflow_template_state_action
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_state_action_delete_view_no_permission(self):
        self._create_test_workflow_template_state_action()
        action_count = self._test_workflow_template_state.actions.count()

        self._clear_events()

        response = self._request_test_worflow_template_state_action_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_action_delete_view_with_access(self):
        self._create_test_workflow_template_state_action()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )
        action_count = self._test_workflow_template_state.actions.count()

        self._clear_events()

        response = self._request_test_worflow_template_state_action_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self._test_workflow_template_state.actions.count(),
            action_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_state_action_edit_view_no_permission(self):
        self._create_test_workflow_template_state_action()
        action_label = self._test_workflow_template_state_action.label

        self._clear_events()

        response = self._request_test_worflow_template_state_action_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_workflow_template_state_action.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state_action.label, action_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_action_edit_view_with_access(self):
        self._create_test_workflow_template_state_action()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )
        action_label = self._test_workflow_template_state_action.label

        self._clear_events()

        response = self._request_test_worflow_template_state_action_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_workflow_template_state_action.refresh_from_db()
        self.assertNotEqual(
            self._test_workflow_template_state_action.label, action_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object,
            self._test_workflow_template_state_action
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_workflow_template)
        self.assertEqual(events[0].verb, event_workflow_template_edited.id)

    def test_workflow_state_action_list_view_no_permission(self):
        self._create_test_workflow_template_state_action()

        self._clear_events()

        response = self._request_test_worflow_template_state_action_list_view()
        self.assertNotContains(
            response=response,
            text=self._test_workflow_template_state_action.label,
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_action_list_view_with_access(self):
        self._create_test_workflow_template_state_action()
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        self._clear_events()

        response = self._request_test_worflow_template_state_action_list_view()
        self.assertContains(
            response=response,
            text=self._test_workflow_template_state_action.label,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_action_selection_view_no_permission(self):
        action_count = self._test_workflow_template_state.actions.count()

        self._clear_events()

        response = self._request_test_workflow_template_state_action_selection_view()
        self.assertEqual(response.status_code, 404)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_workflow_state_action_selection_view_with_access(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )
        action_count = self._test_workflow_template_state.actions.count()

        self._clear_events()

        response = self._request_test_workflow_template_state_action_selection_view()
        self.assertEqual(response.status_code, 302)

        self._test_workflow_template_state.refresh_from_db()
        self.assertEqual(
            self._test_workflow_template_state.actions.count(), action_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_quota_create, icon_quota_delete, icon_quota_edit, icon_quota_list,
    icon_quota_setup
)
from .permissions import (
    permission_quota_create, permission_quota_delete,
    permission_quota_edit, permission_quota_view,
)

link_quota_create = Link(
    icon=icon_quota_create, permissions=(permission_quota_create,),
    text=_('Create quota'), view='quotas:quota_backend_selection'
)
link_quota_delete = Link(
    args='resolved_object.pk', icon=icon_quota_delete,
    permissions=(permission_quota_delete,), tags='dangerous',
    text=_('Delete'), view='quotas:quota_delete'
)
link_quota_edit = Link(
    args='object.pk', icon=icon_quota_edit,
    permissions=(permission_quota_edit,), text=_('Edit'),
    view='quotas:quota_edit'
)
link_quota_list = Link(
    icon=icon_quota_list, text=_('Quotas list'), view='quotas:quota_list'
)
link_quota_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='quotas', model_name='Quota',
        object_permission=permission_quota_view,
        view_permission=permission_quota_create,
    ), icon=icon_quota_setup, text=_('Quotas'),
    view='quotas:quota_list'
)



from mayan.apps.rest_api import generics
from mayan.apps.documents.api_views.api_view_mixins import (
    ParentObjectDocumentAPIViewMixin
)
from mayan.apps.views.generics import DownloadViewMixin

from .permissions import permission_document_file_download


class APIDocumentFileDownloadView(
    DownloadViewMixin, ParentObjectDocumentAPIViewMixin,
    generics.RetrieveAPIView
):
    """
    get: Download a document file.
    """
    lookup_url_kwarg = 'document_file_id'
    mayan_object_permissions = {
        'GET': (permission_document_file_download,),
    }

    def get_download_file_object(self):
        instance = self.get_object()
        instance._event_actor = self.request.user
        return instance.get_download_file_object()

    def get_download_filename(self):
        return self.get_object().filename

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def get_source_queryset(self):
        return self.get_document().files.all()

    def retrieve(self, request, *args, **kwargs):
        return self.render_to_response()



from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        (
            'document_states',
            '0026_alter_workflowtransitiontriggerevent_options'
        )
    ]

    operations = [
        migrations.AddField(
            model_name='workflowinstance', name='datetime',
            field=models.DateTimeField(
                auto_now_add=True, db_index=True,
                default=django.utils.timezone.now,
                help_text='Workflow instance creation date time.',
                verbose_name='Datetime'
            ), preserve_default=False
        )
    ]



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from .models import StoredDriver


@admin.register(StoredDriver)
class StoredDriverAdmin(admin.ModelAdmin):
    list_display = ('internal_name', 'get_label', 'driver_path')

    def get_label(self, instance):
        return instance.driver_label
    get_label.short_description = _('Label')



from django.contrib import admin

from .models import UserMailer


@admin.register(UserMailer)
class UserMailerAdmin(admin.ModelAdmin):
    list_display = (
        'label', 'default', 'enabled', 'backend_path', 'backend_data'
    )



from .literals import (
    TERM_OPERATORS, TERM_MARKER_QUOTE, TERM_MARKER_RAW,
    TERM_MARKER_SPACE_CHARACTER
)


class QueryTerm:
    @staticmethod
    def do_query_parse(query):
        result = {}

        for key, value in query.items():
            result[key] = QueryToken.do_query_parse(value=value)

        return result


class QueryToken:
    @staticmethod
    def do_query_parse(value):
        inside_quotes = False
        inside_raw = False
        result = []
        token_letters = []

        for letter in value:
            if not inside_quotes and letter == TERM_MARKER_RAW:
                if inside_raw:
                    inside_raw = False
                    token_string = ''.join(token_letters)
                    result.append(
                        QueryToken(
                            is_meta=False, is_quoted=inside_quotes,
                            is_raw=True, text=token_string
                        )
                    )
                    token_letters = []
                else:
                    inside_raw = True
            elif not inside_raw and letter == TERM_MARKER_QUOTE:
                if inside_quotes:
                    token_string = ''.join(token_letters)
                    result.append(
                        QueryToken(
                            is_meta=False, is_quoted=True, is_raw=False,
                            text=token_string
                        )
                    )
                    token_letters = []
                    inside_quotes = False
                else:
                    inside_quotes = True
            elif not inside_quotes and not inside_raw and letter == TERM_MARKER_SPACE_CHARACTER:
                if token_letters:
                    token_string = ''.join(token_letters)
                    if token_string in TERM_OPERATORS:
                        is_meta = True
                    else:
                        is_meta = False

                    result.append(
                        QueryToken(
                            is_meta=is_meta, is_quoted=False,
                            is_raw=False, text=token_string
                        )
                    )
                    token_letters = []
            else:
                token_letters.append(letter)

        token_string = ''.join(token_letters)

        if token_string and not inside_quotes and not inside_raw:
            result.append(
                QueryToken(
                    is_meta=False, is_quoted=False, is_raw=False,
                    text=token_string
                )
            )

        return result

    def __init__(self, is_meta, is_quoted, is_raw, text):
        self.is_meta = is_meta
        self.is_quoted = is_quoted
        self.is_raw = is_raw
        self.text = text



import logging

from django.utils.functional import cached_property
from django.utils.translation import ugettext as _

from .exceptions import (
    DynamicSearchBackendException, DynamicSearchScopedQueryError
)
from .literals import (
    ERROR_TEXT_NO_RESULT_SCOPE, SCOPE_DELIMITER,
    SCOPE_MARKER, SCOPE_OPERATOR_CHOICES, SCOPE_RESULT_MARKER,
    TERM_MARKER_QUOTE
)
from .search_query_terms import QueryToken
from .search_query_types import QueryType
from .settings import setting_query_results_limit, setting_results_limit

logger = logging.getLogger(name=__name__)


class ScopedQuery:
    scope_delimiter = SCOPE_DELIMITER
    scope_marker = SCOPE_MARKER

    def __init__(self, search_model):
        self.scope_entry_list = []
        self.search_model = search_model
        self.entry_point = None

    def __str__(self):
        return ', '.join(
            [
                str(scope_entry) for scope_entry in self.scope_entry_list
            ]
        )

    def do_resolve(self, search_backend):
        if self.entry_point:
            id_list = self.entry_point.do_resolve(
                search_backend=search_backend
            )
            if id_list is None:
                raise DynamicSearchScopedQueryError(
                    _(ERROR_TEXT_NO_RESULT_SCOPE)
                )
            else:
                return id_list
        else:
            raise DynamicSearchScopedQueryError(
                _(ERROR_TEXT_NO_RESULT_SCOPE)
            )

    def do_scope_entry_add(self, scope_entry):
        scope_entry.do_pre_add_callback()
        self.scope_entry_list.append(scope_entry)
        scope_entry.do_post_add_callback()

    def do_scope_entry_init(self, **kwargs):
        kwargs['scoped_query'] = self

        scope_entry = ScopedQueryEntry.init(**kwargs)
        self.do_scope_entry_add(scope_entry=scope_entry)

        return scope_entry

    def do_scope_solve(self, search_backend, scope_identifier):
        scope_entry = self.get_scope_entry_by_identifier(
            scope_identifier=scope_identifier
        )

        return scope_entry.do_resolve(search_backend=search_backend)

    def get_scope_entry_by_identifier(self, scope_identifier):
        for scope_entry in self.scope_entry_list:
            scope_entry_identifier = getattr(
                scope_entry, 'scope_identifier', None
            )
            if scope_entry_identifier is not None and scope_entry_identifier == scope_identifier:
                # None is not a valid scope identifier.
                return scope_entry

        raise DynamicSearchScopedQueryError(
            'No scope found with identifier `{}`'.format(scope_identifier)
        )

    def get_scope_identifier_list(self):
        result = []

        for scope_entry in self.scope_entry_list:
            scope_entry_identifier = getattr(
                scope_entry, 'scope_identifier', None
            )
            if scope_entry_identifier:
                result.append(scope_entry_identifier)

        return result

    @property
    def is_empty(self):
        return all(
            scope_entry.is_empty for scope_entry in self.scope_entry_list
        )

    def to_explain(self):
        if self.entry_point:
            return self.entry_point.to_explain()
        else:
            return ''

    def to_string(self):
        scope_text_list = []
        for scope_entry in self.scope_entry_list:
            scope_text_list.append(
                scope_entry.to_string()
            )

        return '&'.join(scope_text_list)


class ScopedQueryEntry:
    _registry = []

    @classmethod
    def _check(cls, key, scoped_query, **kwargs):
        """
        Required subclass class method that determines if the subclass is
        the one that should handle the key and value pair submitted.
        Returns True or False/None.
        """
        raise NotImplementedError

    @classmethod
    def all(cls):
        return cls._registry

    @classmethod
    def check(cls, scoped_query, **kwargs):
        cls.scoped_query = scoped_query
        check_kwargs = cls._check(**kwargs)
        if check_kwargs is not None:
            check_kwargs['scoped_query'] = scoped_query
            return check_kwargs

    @classmethod
    def init(cls, scoped_query, **kwargs):
        """
        Initialization router. Calling this method will cycle all possible
        subclasses and return an instance of the subclass that can handle
        the argument values.
        """
        for klass in cls.all():
            check_kwargs = klass.check(scoped_query=scoped_query, **kwargs)
            if check_kwargs is not None:
                return klass(**check_kwargs)

        raise DynamicSearchScopedQueryError(
            'No query entry found that could handle the arguments: '
            '`{}`, verify the query format, the field name, and the '
            'value data type.'.format(kwargs)
        )

    @classmethod
    def register(cls, klass):
        cls._registry.append(klass)

    def __init__(self, scoped_query, **kwargs):
        self.scoped_query = scoped_query

    def do_post_add_callback(self):
        """
        Optional callback after entry is added to the scoped query.
        """

    def do_pre_add_callback(self):
        """
        Optional callback before entry is added to the scoped query.
        """

    def do_resolve(self, search_backend):
        raise NotImplementedError

    def do_template_render(self, template_name, context=None):
        context_final = self.get_template_context_defaults()

        context_final.update(
            context or {}
        )

        template_string = self.get_template(template_name=template_name)

        return template_string.format(**context_final)

    def get_template(self, template_name):
        return self.templates[template_name]

    def get_template_context_defaults(self):
        return {
            'scope_delimiter': self.scoped_query.scope_delimiter,
            'scope_marker': self.scoped_query.scope_marker
        }

    def get_template_explain_context(self):
        return {}

    def get_template_key_context(self):
        return {}

    def get_template_value_context(self):
        return {}

    @property
    def is_empty(self):
        return False

    def to_explain(self):
        return self.do_template_render(
            template_name='explain',
            context=self.get_template_explain_context()
        )

    def to_string(self):
        return '{key}={value}'.format(
            key=self.do_template_render(
                template_name='key',
                context=self.get_template_key_context()
            ),
            value=self.do_template_render(
                template_name='value',
                context=self.get_template_value_context()
            )
        )


class ScopedQueryEntryData(ScopedQueryEntry):
    def __init__(self, scope_identifier, **kwargs):
        super().__init__(**kwargs)
        self.scope_identifier = scope_identifier

    def do_pre_add_callback(self):
        self.do_scope_identifier_check()

    def do_scope_identifier_check(self):
        try:
            scope_entry = self.scoped_query.get_scope_entry_by_identifier(
                scope_identifier=self.scope_identifier
            )
        except DynamicSearchScopedQueryError:
            # This scope identifier does not exist, it is safe to allow
            # the entry to claim it.
            return
        else:
            raise DynamicSearchScopedQueryError(
                'Scope `{}` is already defined by entry `{}`.'.format(
                    self.scope_identifier, scope_entry.to_string()
                )
            )

    def get_template_key_context(self):
        return {
            'scope_identifier': self.scope_identifier
        }


class ScopedQueryEntryDataFilter(ScopedQueryEntryData):
    templates = {
        'explain': '{field_name} {query_type} {value}',
        'key': '{scope_marker}{scope_identifier}{scope_delimiter}{field_name}',
        'value': '{value}'
    }

    @classmethod
    def _check(cls, key, value):
        if key.startswith(cls.scoped_query.scope_marker):
            # Remove the scope marker.
            key = key[
                len(cls.scoped_query.scope_marker):
            ]

            try:
                scope_identifier, field_name = key.split(
                    cls.scoped_query.scope_delimiter, 1
                )
            except ValueError:
                return None

            any_field = any(
                field_name == search_field_name for search_field_name in cls.scoped_query.search_model.search_field_name_list
            )
            if any_field:
                tokens = QueryToken.do_query_parse(value=value)
                if len(tokens) > 0:
                    if len(tokens) > 1:
                        raise DynamicSearchScopedQueryError(
                            'Each scoped mush only query one term. Multiple '
                            'terms must be split into different scopes. '
                            'To query for multiple words use the symbol '
                            '`{}` to collapase the words into a single '
                            'term.'.format(
                                TERM_MARKER_QUOTE
                            )
                        )
                    else:
                        return {
                            'field_name': field_name,
                            'is_quoted_value': tokens[0].is_quoted,
                            'is_raw_value': tokens[0].is_raw,
                            'scope_identifier': scope_identifier,
                            'value': tokens[0].text
                        }

    def __init__(
        self, field_name, value, is_quoted_value=False, is_raw_value=False,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.field_name = field_name
        self.is_quoted_value = is_quoted_value
        self.is_raw_value = is_raw_value
        self.value = value

    def __str__(self):
        return 'Scope `{}` = query {}: {}'.format(
            self.scope_identifier, self.field_name, self.value
        )

    def do_resolve(self, search_backend):
        if len(
            [
                scope_entry for scope_entry in self.scoped_query.scope_entry_list if isinstance(
                    scope_entry, ScopedQueryEntryDataFilter
                )
            ]
        ) > 1:
            limit = setting_query_results_limit.value
        else:
            limit = setting_results_limit.value

        if self.value:
            query_type, value = QueryType.check_all(value=self.value)

            try:
                results = search_backend._search(
                    limit=limit, is_quoted_value=self.is_quoted_value,
                    is_raw_value=self.is_raw_value, query_type=query_type,
                    search_field=self.search_field, value=value
                )

                if len(results) >= limit:
                    raise DynamicSearchScopedQueryError(
                        _(
                            'Search results exceed limit setting. Results '
                            'might not be reliable if multiple scopes are '
                            'used. Narrow down the search criteria or '
                            'increase the value of the results limit '
                            'setting.'
                        )
                    )

                return results
            except DynamicSearchBackendException:
                """Raise `DynamicSearchBackendException` as is."""
                raise
            except Exception as exception:
                """Wrap any other exception raised by the backend."""
                raise DynamicSearchBackendException(
                    _(
                        'Search backend error. Verify that the search '
                        'service is available and that the search syntax '
                        'is valid for the active search backend; %s'
                    ) % exception
                ) from exception
        else:
            return ()

    def get_template_explain_context(self):
        context = super().get_template_explain_context()

        query_type, value = QueryType.check_all(value=self.value)

        if self.is_quoted_value:
            value = '"{}"'.format(value)
        elif self.is_raw_value:
            value = '`{}`'.format(value)

        context.update(
            {
                'field_name': self.field_name,
                'query_type': query_type.name,
                'value': value
            }
        )
        return context

    def get_template_key_context(self):
        context = super().get_template_key_context()

        context.update(
            {'field_name': self.field_name}
        )

        return context

    def get_template_value_context(self):
        return {
            'value': self.value
        }

    @property
    def is_empty(self):
        return not self.value and not self.is_quoted_value and not self.is_raw_value

    @cached_property
    def search_field(self):
        return self.scoped_query.search_model.get_search_field(
            field_name=self.field_name
        )


class ScopedQueryEntryDataOperator(ScopedQueryEntryData):
    templates = {
        'explain': '{operator_with_operands}',
        'key': '{scope_marker}{scope_identifier}',
        'value': '{operator_text}{scope_delimiter}{operand_list}'
    }

    @classmethod
    def _check(cls, key, value):
        if key.startswith(cls.scoped_query.scope_marker):
            scope_identifier = key[len(cls.scoped_query.scope_marker):]

            # Check for operator text in value.
            # Format: `SCOPE`=`OPERATOR`_`SOURCE_SCOPES`_...
            parts = value.split(cls.scoped_query.scope_delimiter)

            # At least three parts are needed: operator, two operands.
            if len(parts) >= 3:
                if any(
                    operator_text == parts[0] for operator_text in SCOPE_OPERATOR_CHOICES
                ):
                    return {
                        'operand_list': parts[1:],
                        'operator_text': parts[0],
                        'scope_identifier': scope_identifier
                    }

    def __init__(self, operand_list, operator_text, **kwargs):
        super().__init__(**kwargs)

        if len(operand_list) < 2:
            raise DynamicSearchScopedQueryError(
                'Invalid operator declaration. Must specify '
                'at least two source scopes.'
            )

        self.operand_list = operand_list
        self.operator_text = operator_text

        if self.scope_identifier in self.operand_list:
            raise DynamicSearchScopedQueryError(
                'Operator scope `{}` is trying to resolve an operand '
                'with the same scope identifier. Self references are '
                'invalid. Verify the scope identifiers '
                'of the query.'.format(self.scope_identifier)
            )

    def __str__(self):
        return 'Scope `{}` = operator {}: {}'.format(
            self.scope_identifier, self.operator_text, self.operand_list
        )

    def do_resolve(self, search_backend):
        result = None

        for scope_child in self.operand_list:
            id_list = self.scoped_query.do_scope_solve(
                scope_identifier=scope_child,
                search_backend=search_backend
            )

            if result is None:
                result = id_list
            else:
                result = self.operator_function(result, id_list)

        return result

    def get_template_explain_context(self):
        context = super().get_template_explain_context()

        result = []

        for operand in self.operand_list:
            scope_entry = self.scoped_query.get_scope_entry_by_identifier(scope_identifier=operand)
            result.append(scope_entry.to_explain())

        operator_with_operands = ' {} '.format(self.operator_text).join(result)

        context.update(
            {
                'operator_with_operands': '( {} )'.format(
                    operator_with_operands
                )
            }
        )
        return context

    def get_template_key_context(self):
        context = super().get_template_key_context()

        context.update(
            {
                'scope_identifier': self.scope_identifier
            }
        )

        return context

    def get_template_value_context(self):
        return {
            'operator_text': self.operator_text,
            'operand_list': self.scoped_query.scope_delimiter.join(
                self.operand_list
            )
        }

    @property
    def operator_function(self):
        return SCOPE_OPERATOR_CHOICES.get(self.operator_text)


class ScopedQueryEntryControl(ScopedQueryEntry):
    """Base class for entries that control the flow of the scoped query."""


class ScopedQueryEntryControlResult(ScopedQueryEntryControl):
    templates = {
        'explain': '{scope_entry}',
        'key': '{scope_marker}{scope_result_marker}',
        'value': '{result_scope_identifier}'
    }

    @classmethod
    def _check(cls, key, value):
        if key == '{}{}'.format(
            cls.scoped_query.scope_marker, SCOPE_RESULT_MARKER
        ):
            return {
                'result_scope_identifier': value
            }

    def __init__(self, result_scope_identifier, **kwargs):
        super().__init__(**kwargs)
        self.result_scope_identifier = result_scope_identifier

    def __str__(self):
        return 'Result: {}'.format(self.result_scope_identifier)

    def do_post_add_callback(self):
        self.scoped_query.entry_point = self

    def do_resolve(self, search_backend):
        if self.result_scope_identifier:
            return self.scoped_query.do_scope_solve(
                scope_identifier=self.result_scope_identifier,
                search_backend=search_backend
            )

    def get_template_explain_context(self):
        context = super().get_template_explain_context()

        scope_entry = self.scoped_query.get_scope_entry_by_identifier(
            scope_identifier=self.result_scope_identifier
        )

        context.update(
            {
                'scope_entry': scope_entry.to_explain()
            }
        )

        return context

    def get_template_key_context(self):
        context = super().get_template_key_context()

        context.update(
            {
                'scope_result_marker': SCOPE_RESULT_MARKER
            }
        )

        return context

    def get_template_value_context(self):
        context = super().get_template_value_context()

        context.update(
            {
                'result_scope_identifier': self.result_scope_identifier,
                'scope_result_marker': SCOPE_RESULT_MARKER
            }
        )

        return context


ScopedQueryEntry.register(klass=ScopedQueryEntryDataOperator)
ScopedQueryEntry.register(klass=ScopedQueryEntryDataFilter)
ScopedQueryEntry.register(klass=ScopedQueryEntryControlResult)



from django.apps import apps

from .literals import (
    DEFAULT_DOCUMENT_TYPE_LABEL, STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE,
    STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE
)
from .settings import (
    setting_document_file_page_image_cache_maximum_size,
    setting_document_version_page_image_cache_maximum_size
)
from .signals import signal_post_initial_document_type


def handler_create_default_document_type(sender, **kwargs):
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )

    if not DocumentType.objects.count():
        document_type = DocumentType.objects.create(
            label=DEFAULT_DOCUMENT_TYPE_LABEL
        )
        signal_post_initial_document_type.send(
            sender=DocumentType, instance=document_type
        )


def handler_create_document_file_page_image_cache(sender, **kwargs):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')
    Cache.objects.update_or_create(
        defaults={
            'maximum_size': setting_document_file_page_image_cache_maximum_size.value,
        }, defined_storage_name=STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE
    )


def handler_create_document_version_page_image_cache(sender, **kwargs):
    Cache = apps.get_model(app_label='file_caching', model_name='Cache')
    Cache.objects.update_or_create(
        defaults={
            'maximum_size': setting_document_version_page_image_cache_maximum_size.value,
        },
        defined_storage_name=STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE
    )



from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

from mayan.apps.rest_api import serializers


class ContentTypeSerializer(serializers.ModelSerializer):
    class Meta:
        extra_kwargs = {
            'url': {
                'label': _('URL'),
                'lookup_url_kwarg': 'content_type_id',
                'view_name': 'rest_api:content_type-detail'
            }
        }
        fields = ('app_label', 'id', 'model', 'url')
        model = ContentType
        read_only_fields = ('app_label', 'id', 'model')



from django.utils.translation import ugettext_lazy as _

from mayan.apps.rest_api import serializers

from .models import Asset


class AppImageErrorSerializer(serializers.Serializer):
    name = serializers.CharField(
        label=_('Name')
    )
    image_path = serializers.CharField(
        label=_('Image path')
    )
    image_url = serializers.HyperlinkedIdentityField(
        label=_('Image URL'), lookup_field='name',
        lookup_url_kwarg='app_image_error_name',
        view_name='rest_api:app_image_error-image'
    )
    template_name = serializers.CharField(
        label=_('Template name')
    )
    url = serializers.HyperlinkedIdentityField(
        label=_('URL'), lookup_field='name',
        lookup_url_kwarg='app_image_error_name',
        view_name='rest_api:app_image_error-detail'
    )


class AssetSerializer(serializers.HyperlinkedModelSerializer):
    image_url = serializers.HyperlinkedIdentityField(
        label=_('Image URL'), lookup_url_kwarg='asset_id',
        view_name='rest_api:asset-image'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _('URL'),
                'lookup_url_kwarg': 'asset_id',
                'view_name': 'rest_api:asset-detail'
            }
        }
        fields = (
            'file', 'label', 'id', 'image_url', 'internal_name', 'url'
        )
        model = Asset



import logging

from django.apps import apps
from django.db.models.signals import post_save, post_delete
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_list_facet, menu_object, menu_secondary, menu_tools
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn

from .events import (
    event_detached_signature_created, event_detached_signature_deleted,
    event_detached_signature_uploaded, event_embedded_signature_created
)
from .handlers import (
    handler_unverify_key_signatures, handler_verify_key_signatures
)
from .hooks import (
    hook_create_embedded_signature, hook_decrypt_document_file
)
from .links import (
    link_document_file_all_signature_refresh,
    link_document_file_all_signature_verify,
    link_document_file_signature_detached_delete,
    link_document_file_signature_detached_create,
    link_document_file_signature_embedded_create,
    link_document_file_signature_detail,
    link_document_file_signature_detached_download,
    link_document_file_signature_list,
    link_document_file_signature_detached_upload,
)
from .permissions import (
    permission_document_file_sign_detached,
    permission_document_file_sign_embedded,
    permission_document_file_signature_delete,
    permission_document_file_signature_download,
    permission_document_file_signature_upload,
    permission_document_file_signature_view,
)

logger = logging.getLogger(name=__name__)


class DocumentSignaturesApp(MayanAppConfig):
    app_namespace = 'signatures'
    app_url = 'signatures'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.document_signatures'
    verbose_name = _('Document signatures')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )

        Key = apps.get_model(
            app_label='django_gpg', model_name='Key'
        )

        DetachedSignature = self.get_model(model_name='DetachedSignature')
        EmbeddedSignature = self.get_model(model_name='EmbeddedSignature')
        SignatureBaseModel = self.get_model(model_name='SignatureBaseModel')

        DocumentFile.register_post_save_hook(
            func=hook_create_embedded_signature, order=1
        )
        DocumentFile.register_pre_open_hook(
            func=hook_decrypt_document_file, order=1
        )

        EventModelRegistry.register(model=DetachedSignature)
        EventModelRegistry.register(model=EmbeddedSignature)

        ModelEventType.register(
            model=DocumentFile, event_types=(
                event_detached_signature_created,
                event_detached_signature_deleted,
                event_detached_signature_uploaded,
                event_embedded_signature_created
            )
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_document_file_sign_detached,
                permission_document_file_sign_embedded,
                permission_document_file_signature_delete,
                permission_document_file_signature_download,
                permission_document_file_signature_view,
                permission_document_file_signature_upload,
            )
        )
        ModelPermission.register_inheritance(
            model=SignatureBaseModel, related='document_file'
        )
        ModelPermission.register_inheritance(
            model=DetachedSignature, related='document_file'
        )

        SourceColumn(
            attribute='date_time', label=_('Date and time'),
            source=SignatureBaseModel
        )
        SourceColumn(
            attribute='get_key_id', label=_('Key ID'),
            source=SignatureBaseModel
        )
        SourceColumn(
            func=lambda context: context['object'].signature_id or _('None'),
            label=_('Signature ID'), source=SignatureBaseModel
        )
        SourceColumn(
            func=lambda context: SignatureBaseModel.objects.get_subclass(
                pk=context['object'].pk
            ).get_signature_type_display(), label=_('Type'),
            source=SignatureBaseModel
        )

        # Document file

        menu_list_facet.bind_links(
            links=(link_document_file_signature_list,),
            sources=(DocumentFile,)
        )

        # Signatures

        menu_object.bind_links(
            links=(
                link_document_file_signature_detail,
                link_document_file_signature_detached_download,
                link_document_file_signature_detached_delete,
            ), sources=(SignatureBaseModel,)
        )

        menu_secondary.bind_links(
            links=(
                link_document_file_signature_detached_create,
                link_document_file_signature_embedded_create,
                link_document_file_signature_detached_upload
            ), sources=(
                'signatures:document_file_signature_list',
                'signatures:document_file_signature_detached_create',
                'signatures:document_file_signature_upload',
                'signatures:document_file_signature_embedded_create',
                'signatures:document_file_signature_detached_delete',
                'signatures:document_file_signature_detail',
                'signatures:document_file_signature_detached_download',
            )
        )

        # Tools

        menu_tools.bind_links(
            links=(
                link_document_file_all_signature_refresh,
                link_document_file_all_signature_verify,
            )
        )

        post_delete.connect(
            dispatch_uid='signatures_handler_unverify_key_signatures',
            receiver=handler_unverify_key_signatures,
            sender=Key
        )
        post_save.connect(
            dispatch_uid='signatures_handler_verify_key_signatures',
            receiver=handler_verify_key_signatures,
            sender=Key
        )



from mayan.apps.rest_api import generics

from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)
from ..serializers.workflow_template_transition_serializers import (
    WorkflowTemplateTransitionSerializer, WorkflowTransitionFieldSerializer,
    WorkflowTemplateTransitionTriggerSerializer
)

from .api_view_mixins import (
    ParentObjectWorkflowTemplateAPIViewMixin,
    ParentObjectWorkflowTemplateTransitionAPIViewMixin
)


# Workflow transition views


class APIWorkflowTemplateTransitionListView(
    ParentObjectWorkflowTemplateAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the workflow template transitions.
    post: Create a new workflow template transition.
    """
    mayan_external_object_permissions = {
        'GET': (permission_workflow_template_view,),
        'POST': (permission_workflow_template_edit,)
    }
    ordering_fields = ('destination_state', 'id', 'label', 'origin_state')
    serializer_class = WorkflowTemplateTransitionSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'workflow': self.get_workflow_template()
        }

    def get_source_queryset(self):
        return self.get_workflow_template().transitions.all()


class APIWorkflowTemplateTransitionDetailView(
    ParentObjectWorkflowTemplateAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template transition.
    get: Return the details of the selected workflow template transition.
    patch: Edit the selected workflow template transition.
    put: Edit the selected workflow template transition.
    """
    mayan_object_permissions = {
        'DELETE': (permission_workflow_template_edit,),
        'GET': (permission_workflow_template_view,),
        'PATCH': (permission_workflow_template_edit,),
        'PUT': (permission_workflow_template_edit,)
    }
    lookup_url_kwarg = 'workflow_template_transition_id'
    serializer_class = WorkflowTemplateTransitionSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'workflow': self.get_workflow_template()
        }

    def get_source_queryset(self):
        return self.get_workflow_template().transitions.all()


# Workflow template transition fields


class APIWorkflowTemplateTransitionFieldListView(
    ParentObjectWorkflowTemplateTransitionAPIViewMixin,
    generics.ListCreateAPIView
):
    """
    get: Returns a list of all the workflow template transition fields.
    post: Create a new workflow template transition field.
    """
    mayan_external_object_permissions = {
        'GET': (permission_workflow_template_view,),
        'POST': (permission_workflow_template_edit,)
    }
    ordering_fields = ('id', 'label', 'name', 'required', 'widget_kwargs')
    serializer_class = WorkflowTransitionFieldSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'transition': self.get_workflow_template_transition()
        }

    def get_source_queryset(self):
        return self.get_workflow_template_transition().fields.all()


class APIWorkflowTemplateTransitionFieldDetailView(
    ParentObjectWorkflowTemplateTransitionAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template transition field.
    get: Return the details of the selected workflow template transition field.
    patch: Edit the selected workflow template transition field.
    put: Edit the selected workflow template transition field.
    """
    mayan_object_permissions = {
        'DELETE': (permission_workflow_template_edit,),
        'GET': (permission_workflow_template_view,),
        'PATCH': (permission_workflow_template_edit,),
        'PUT': (permission_workflow_template_edit,)
    }
    lookup_url_kwarg = 'workflow_template_transition_field_id'
    serializer_class = WorkflowTransitionFieldSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_source_queryset(self):
        return self.get_workflow_template_transition().fields.all()


# Workflow template transition triggers


class APIWorkflowTemplateTransitionTriggerListView(
    ParentObjectWorkflowTemplateTransitionAPIViewMixin,
    generics.ListCreateAPIView
):
    """
    get: Returns a list of all the workflow template transition triggers.
    post: Create a new workflow template transition trigger.
    """
    mayan_external_object_permissions = {
        'GET': (permission_workflow_template_view,),
        'POST': (permission_workflow_template_edit,)
    }
    ordering_fields = ('event_type', 'id')
    serializer_class = WorkflowTemplateTransitionTriggerSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'transition': self.get_workflow_template_transition()
        }

    def get_source_queryset(self):
        return self.get_workflow_template_transition().trigger_events.all()


class APIWorkflowTemplateTransitionTriggerDetailView(
    ParentObjectWorkflowTemplateTransitionAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template transition trigger.
    get: Return the details of the selected workflow template transition trigger.
    patch: Edit the selected workflow template transition trigger.
    put: Edit the selected workflow template transition trigger.
    """
    mayan_object_permissions = {
        'DELETE': (permission_workflow_template_edit,),
        'GET': (permission_workflow_template_view,),
        'PATCH': (permission_workflow_template_edit,),
        'PUT': (permission_workflow_template_edit,)
    }
    lookup_url_kwarg = 'workflow_template_transition_trigger_id'
    serializer_class = WorkflowTemplateTransitionTriggerSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_source_queryset(self):
        return self.get_workflow_template_transition().trigger_events.all()



from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from ...models import UserOTPData


class Command(BaseCommand):
    help = 'Initialize the OTP state data for all users.'

    def handle(self, *args, **options):
        count = 0

        for user in get_user_model().objects.all():
            user_data, created = UserOTPData.objects.get_or_create(user=user)
            if created:
                count += 1

        self.stdout.write(msg='Completed for {} users.'.format(count))



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('acls', '0003_auto_20180402_0339'),
    ]

    operations = [
        migrations.CreateModel(
            name='GlobalAccessControlListProxy',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('acls.accesscontrollist',),
        ),
        migrations.AlterField(
            model_name='accesscontrollist',
            name='role',
            field=models.ForeignKey(help_text='Role to which the access is granted for the specified object.', on_delete=django.db.models.deletion.CASCADE, related_name='acls', to='permissions.Role', verbose_name='Role'),
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.sources.source_backend_actions.interface_arguments import SourceBackendActionInterfaceArgument

argument_expand = SourceBackendActionInterfaceArgument(
    default=False, help_text=_(
        'Controls whether or not the uploaded file will be uncompressed '
        'and processed and individual files.'
    ), required=False
)



from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter

from .events import event_credential_used


class StoredCredentialBusinessLogicMixin:
    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_credential_used,
        target='self'
    )
    def get_backend_data(self):
        obj = super().get_backend_data()
        backend_class = self.get_backend_class()

        if hasattr(backend_class, 'post_processing'):
            obj = backend_class.post_processing(obj=obj)

        return obj



import logging

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.utils.deprecation import MiddlewareMixin

logger = logging.getLogger(name=__name__)


class ErrorLoggingMiddleware(MiddlewareMixin):
    def process_exception(self, request, exception):
        if (settings.TESTING and not isinstance(exception, (PermissionDenied, Http404))) or not settings.TESTING:
            # Don't log non critical exceptions in testing mode.
            logger.exception(
                'Exception caught by request middleware; %s, %s', request,
                exception
            )



import logging
import os
from shutil import copyfileobj
import subprocess

from django.apps import apps
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.converter.classes import ConverterBase
from mayan.apps.converter.literals import CONVERTER_OFFICE_FILE_MIMETYPES
from mayan.apps.storage.utils import NamedTemporaryFile

from .exceptions import ParserError
from .settings import setting_pdftotext_path

logger = logging.getLogger(name=__name__)


class Parser:
    """
    Parser base class.
    """
    _registry = {}

    @classmethod
    def parse_document_file_page(cls, document_file_page):
        parser_classes = cls._registry.get(
            document_file_page.document_file.mimetype, ()
        )

        for parser_class in parser_classes:
            try:
                parser = parser_class()
                parser.process_document_file_page(
                    document_file_page=document_file_page
                )
            except ParserError:
                """If parser raises error, try next parser in the list."""
            else:
                # If parser was successfull there is no need to try
                # others in the list for this mimetype.
                return

    @classmethod
    def parse_document_file(cls, document_file):
        parser_classes = cls._registry.get(
            document_file.mimetype, ()
        )

        for parser_class in parser_classes:
            try:
                parser = parser_class()
                parser.process_document_file(document_file)
            except ParserError:
                """If parser raises error, try next parser in the list."""
            else:
                # If parser was successfull there is no need to try
                # others in the list for this mimetype.
                return

    @classmethod
    def register(cls, mimetypes, parser_classes):
        for mimetype in mimetypes:
            for parser_class in parser_classes:
                cls._registry.setdefault(
                    mimetype, []
                ).append(parser_class)

    def process_document_file(self, document_file):
        logger.info(
            'Starting parsing for document file: %s', document_file
        )
        logger.debug('document file: %d', document_file.pk)

        for document_file_page in document_file.pages.all():
            self.process_document_file_page(
                document_file_page=document_file_page
            )

    def process_document_file_page(self, document_file_page):
        DocumentFilePageContent = apps.get_model(
            app_label='document_parsing',
            model_name='DocumentFilePageContent'
        )

        logger.info(
            'Processing page: %d of document file: %s',
            document_file_page.page_number, document_file_page.document_file
        )

        with document_file_page.document_file.open() as file_object:
            try:
                parsed_content = self.execute(
                    file_object=file_object,
                    page_number=document_file_page.page_number
                )

                DocumentFilePageContent.objects.update_or_create(
                    document_file_page=document_file_page, defaults={
                        'content': parsed_content
                    }
                )

            except Exception as exception:
                error_message = _('Exception parsing page; %s') % exception
                logger.error(error_message, exc_info=True)
                raise ParserError(error_message)
            finally:
                file_object.close()

        logger.info(
            'Finished processing page: %d of document file: %s',
            document_file_page.page_number, document_file_page.document_file
        )

    def execute(self, file_object, page_number):
        raise NotImplementedError(
            'Your %s class has not defined the required execute() method.' %
            self.__class__.__name__
        )


class PopplerParser(Parser):
    """
    PDF parser using the pdftotext execute from the poppler package.
    """
    def __init__(self):
        self.pdftotext_path = setting_pdftotext_path.value
        if not os.path.exists(self.pdftotext_path):
            error_message = _(
                'Cannot find pdftotext executable at: %s'
            ) % self.pdftotext_path
            logger.error(error_message)
            raise ParserError(error_message)

        logger.debug('self.pdftotext_path: %s', self.pdftotext_path)

    def execute(self, file_object, page_number):
        logger.debug('Parsing PDF page: %d', page_number)

        with NamedTemporaryFile() as temporary_file_object:
            copyfileobj(fsrc=file_object, fdst=temporary_file_object)
            temporary_file_object.seek(0)

            command = []
            command.append(self.pdftotext_path)
            command.append('-f')
            command.append(
                str(page_number)
            )
            command.append('-l')
            command.append(
                str(page_number)
            )
            command.append(temporary_file_object.name)
            command.append('-')

            proc = subprocess.Popen(
                command, close_fds=True, stderr=subprocess.PIPE,
                stdout=subprocess.PIPE
            )
            return_code = proc.wait()
            if return_code != 0:
                logger.error(
                    proc.stderr.readline()
                )

                raise ParserError

            output = proc.stdout.read()

            if output == b'\x0c':
                logger.debug('Parser didn\'t return any output')
                return ''

            if output[-3:] == b'\x0a\x0a\x0c':
                return force_text(
                    s=output[:-3]
                )

            return force_text(s=output)


class OfficePopplerParser(PopplerParser):
    def execute(self, file_object, page_number):
        converter = ConverterBase.get_converter_class()(
            file_object=file_object
        )
        with converter.to_pdf() as pdf_file_object:
            return super().execute(
                file_object=pdf_file_object, page_number=page_number
            )


Parser.register(
    mimetypes=('application/pdf',),
    parser_classes=(PopplerParser,)
)
Parser.register(
    mimetypes=CONVERTER_OFFICE_FILE_MIMETYPES,
    parser_classes=(OfficePopplerParser,)
)



import io
import logging
import shutil

from django.db import models

from mayan.apps.storage.utils import NamedTemporaryFile

from .classes import GPGBackend, KeyStub, SignatureVerification
from .exceptions import (
    DecryptionError, KeyDoesNotExist, KeyFetchingError, VerificationError
)
from .literals import KEY_TYPE_PUBLIC, KEY_TYPE_SECRET
from .settings import setting_keyserver

logger = logging.getLogger(name=__name__)


class KeyManager(models.Manager):
    def _preload_keys(
        self, all_keys=False, key_fingerprint=None, key_id=None
    ):
        # Preload keys.
        if all_keys:
            logger.debug(msg='preloading all keys')
            keys = self.values()
        elif key_fingerprint:
            logger.debug('preloading key fingerprint: %s', key_fingerprint)
            keys = self.filter(fingerprint=key_fingerprint).values()
            if not keys:
                logger.debug(
                    'key fingerprint %s not found', key_fingerprint
                )
                raise KeyDoesNotExist(
                    'Specified key for verification not found'
                )
        elif key_id:
            logger.debug('preloading key id: %s', key_id)
            keys = self.filter(fingerprint__endswith=key_id).values()
            if keys:
                logger.debug('key id %s impored', key_id)
            else:
                logger.debug('key id %s not found', key_id)
        else:
            keys = ()

        return keys

    def decrypt_file(
        self, file_object, all_keys=False, key_fingerprint=None, key_id=None
    ):
        keys = self._preload_keys(
            all_keys=all_keys, key_fingerprint=key_fingerprint, key_id=key_id
        )

        decrypt_result = GPGBackend.get_instance().decrypt_file(
            file_object=file_object, keys=keys
        )

        logger.debug('decrypt_result.status: %s', decrypt_result.status)

        if not decrypt_result.status or decrypt_result.status == 'no data was provided':
            raise DecryptionError('Unable to decrypt file')

        file_object.close()

        return io.BytesIO(initial_bytes=decrypt_result.data)

    def private_keys(self):
        return self.filter(key_type=KEY_TYPE_SECRET)

    def public_keys(self):
        return self.filter(key_type=KEY_TYPE_PUBLIC)

    def receive_key(self, key_id):
        key_data = GPGBackend.get_instance().recv_keys(
            key_id=key_id, keyserver=setting_keyserver.value
        )

        if not key_data:
            raise KeyFetchingError('No key found')
        else:
            return self.create(key_data=key_data)

    def search(self, query):
        key_data_list = GPGBackend.get_instance().search_keys(
            keyserver=setting_keyserver.value, query=query
        )

        result = []
        for key_data in key_data_list:
            result.append(KeyStub(raw=key_data))

        return result

    def verify_file(
        self, file_object, signature_file=None, all_keys=False,
        key_fingerprint=None, key_id=None
    ):
        keys = self._preload_keys(
            all_keys=all_keys, key_fingerprint=key_fingerprint,
            key_id=key_id
        )

        if signature_file:
            # Save the original data and invert the argument order:
            # signature first, file second.
            with NamedTemporaryFile() as temporary_file_object:
                shutil.copyfileobj(
                    fsrc=file_object, fdst=temporary_file_object
                )
                temporary_file_object.seek(0)

                with NamedTemporaryFile() as temporary_signature_file_object:
                    shutil.copyfileobj(
                        fsrc=signature_file,
                        fdst=temporary_signature_file_object
                    )
                    temporary_signature_file_object.seek(0)
                    signature_file.seek(0)
                    verify_result = GPGBackend.get_instance().verify_file(
                        data_filename=temporary_file_object.name,
                        file_object=temporary_signature_file_object,
                        keys=keys
                    )
        else:
            verify_result = GPGBackend.get_instance().verify_file(
                file_object=file_object, keys=keys
            )

        logger.debug('verify_result.status: %s', verify_result.status)

        if verify_result:
            # Signed and key present.
            logger.debug(msg='signed and key present')
            return SignatureVerification(verify_result.__dict__)
        elif verify_result.status == 'no public key' and not (key_fingerprint or all_keys or key_id):
            # Signed but key not present, retry with key fetch.
            logger.debug(msg='no public key')
            file_object.seek(0)
            return self.verify_file(
                file_object=file_object, signature_file=signature_file,
                key_id=verify_result.key_id
            )
        elif verify_result.key_id:
            # Signed, retried and key still not found.
            logger.debug(msg='signed, retried and key still not found')
            return SignatureVerification(verify_result.__dict__)
        else:
            logger.debug(msg='file not signed')
            raise VerificationError('File not signed')



import re

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion

import mayan.apps.databases.model_mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('documents', '0080_populate_file_size')
    ]

    operations = [
        migrations.CreateModel(
            name='SignatureCapture',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('data', models.TextField(blank=True, help_text='Data representing the handwritten signature.', verbose_name='Signature capture data')),
                ('svg', models.TextField(blank=True, help_text='Vector representation of the handwritten signature.', verbose_name='SVG signature capture data')),
                ('text', models.CharField(help_text='Print version of the captured signature.', max_length=224, verbose_name='Text')),
                ('date_time_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Date and time created')),
                ('date_time_edited', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Date and time edited')),
                ('internal_name', models.CharField(db_index=True, help_text='This value will be used when referencing this signature capture in relationship to the document. Can only contain letters, numbers, and underscores.', max_length=255, validators=[django.core.validators.RegexValidator(re.compile('^[a-zA-Z0-9_]+\\Z'), "Enter a valid 'internal name' consisting of letters, numbers, and underscores.", 'invalid')], verbose_name='Internal name')),
                ('document', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='signature_captures', to='documents.document', verbose_name='Document')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='signature_captures', to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Signature capture',
                'verbose_name_plural': 'Signature captures',
                'ordering': ('-date_time_created',),
                'unique_together': {('document', 'internal_name')}
            },
            bases=(mayan.apps.databases.model_mixins.ExtraDataModelMixin, models.Model)
        )
    ]



class TrashedDocumentAPIViewTestMixin:
    def _request_test_document_trash_api_view(self):
        return self.delete(
            viewname='rest_api:document-detail', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_trashed_document_delete_api_view(self):
        return self.delete(
            viewname='rest_api:trasheddocument-detail', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_trashed_document_detail_api_view(self):
        return self.get(
            viewname='rest_api:trasheddocument-detail', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_trashed_document_image_api_view(self):
        return self.get(
            viewname='rest_api:trasheddocument-image', kwargs={
                'document_id': self._test_document.pk,
            }
        )

    def _request_test_trashed_document_list_api_view(self):
        return self.get(
            viewname='rest_api:trasheddocument-list'
        )

    def _request_test_trashed_document_restore_via_get_api_view(self):
        return self.get(
            viewname='rest_api:trasheddocument-restore', kwargs={
                'document_id': self._test_document.pk
            }, query={'format': 'api'}
        )

    def _request_test_trashed_document_restore_via_post_api_view(self):
        return self.post(
            viewname='rest_api:trasheddocument-restore', kwargs={
                'document_id': self._test_document.pk
            }
        )


class TrashedDocumentViewTestMixin:
    def _request_test_document_trash_get_view(self):
        return self.get(
            viewname='documents:document_trash', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_trash_post_view(self):
        return self.post(
            viewname='documents:document_trash', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_empty_trash_view(self):
        return self.post(viewname='documents:trash_can_empty')

    def _request_test_trashed_document_restore_get_view(self):
        return self.get(
            viewname='documents:document_restore', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_trashed_document_restore_post_view(self):
        return self.post(
            viewname='documents:document_restore', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_trashed_document_delete_get_view(self):
        return self.get(
            viewname='documents:document_delete', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_trashed_document_delete_post_view(self):
        return self.post(
            viewname='documents:document_delete', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_trashed_document_list_view(self):
        return self.get(viewname='documents:document_list_deleted')



from datetime import datetime

from django.db import migrations
from django.utils.encoding import force_text
from django.utils.timezone import make_aware

from ..classes import GPGBackend


def code_save_keys(apps, schema_editor):
    # Refill the creation and expiration fields with a date and time value.
    Key = apps.get_model(
        app_label='django_gpg', model_name='Key'
    )

    for key in Key.objects.using(alias=schema_editor.connection.alias).all():
        key_data = force_text(s=key.key_data)
        import_results, key_info = GPGBackend.get_instance().import_and_list_keys(
            key_data=key_data
        )

        key.creation_date = make_aware(
            value=datetime.fromtimestamp(int(key_info['date']))
        )
        if key_info['expires']:
            key.expiration_date = make_aware(
                value=datetime.fromtimestamp(
                    int(key_info['expires'])
                )
            )
        key.save()


def code_save_keys_reverse(apps, schema_editor):
    # Remove the time component of the creation and expiration fields.
    Key = apps.get_model(
        app_label='django_gpg', model_name='Key'
    )

    for key in Key.objects.using(alias=schema_editor.connection.alias).all():
        key_data = force_text(s=key.key_data)
        import_results, key_info = GPGBackend.get_instance().import_and_list_keys(
            key_data=key_data
        )

        key.creation_date = make_aware(
            value=datetime.fromtimestamp(int(key_info['date']))
        ).date()
        if key_info['expires']:
            key.expiration_date = make_aware(
                value=datetime.fromtimestamp(
                    int(key_info['expires'])
                )
            ).date()
        key.save()


class Migration(migrations.Migration):
    """
    Known issue: Reverting this migration causes the key access to return the
    error:
    `invalid literal for int() with base 10: b'24 00:00:00'`
    as the date field tries to parse the unused time string fragment.
    There are no known solutions.
    Only workarounds is to execute the .save() method for each key to save
    the proper format value for the reverted date fields.
    2021-01-28 Django version 2.2.16. Mayan EDMS version 4.0.
    """
    dependencies = [
        ('django_gpg', '0007_auto_20210128_0504')
    ]

    operations = [
        migrations.RunPython(
            code=code_save_keys,
            reverse_code=code_save_keys_reverse
        )
    ]



from django.contrib import messages
from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventType
from mayan.apps.events.models import StoredEventType
from mayan.apps.views.generics import (
    FormView, SingleObjectCreateView, SingleObjectDeleteView,
    SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms import (
    WorkflowTransitionForm, WorkflowTransitionTriggerEventRelationshipFormSet
)
from ..icons import (
    icon_workflow_template_transition,
    icon_workflow_template_transition_create,
    icon_workflow_template_transition_delete,
    icon_workflow_template_transition_edit,
    icon_workflow_template_transition_list,
    icon_workflow_template_transition_triggers,
    icon_workflow_template_transition_field,
    icon_workflow_template_transition_field_create,
    icon_workflow_template_transition_field_delete,
    icon_workflow_template_transition_field_edit,
    icon_workflow_template_transition_field_list
)
from ..links import (
    link_workflow_template_transition_create,
    link_workflow_template_transition_field_create
)
from ..models import Workflow, WorkflowTransition, WorkflowTransitionField
from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)


class WorkflowTemplateTransitionCreateView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_class = Workflow
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_id'
    form_class = WorkflowTransitionForm
    view_icon = icon_workflow_template_transition_create

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                'Create transitions for workflow: %s'
            ) % self.external_object,
            'workflow': self.external_object
        }

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['workflow'] = self.external_object
        return kwargs

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'workflow': self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.transitions.all()

    def get_success_url(self):
        return reverse(
            viewname='document_states:workflow_template_transition_list',
            kwargs={
                'workflow_template_id': self.kwargs[
                    'workflow_template_id'
                ]
            }
        )


class WorkflowTemplateTransitionDeleteView(SingleObjectDeleteView):
    model = WorkflowTransition
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_transition_id'
    view_icon = icon_workflow_template_transition_delete

    def get_extra_context(self):
        return {
            'object': self.object,
            'navigation_object_list': ('object', 'workflow'),
            'title': _(
                'Delete workflow transition: %s?'
            ) % self.object,
            'workflow': self.object.workflow
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_success_url(self):
        return reverse(
            viewname='document_states:workflow_template_transition_list',
            kwargs={'workflow_template_id': self.object.workflow.pk}
        )


class WorkflowTemplateTransitionEditView(SingleObjectEditView):
    form_class = WorkflowTransitionForm
    model = WorkflowTransition
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_transition_id'
    view_icon = icon_workflow_template_transition_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow'),
            'object': self.object,
            'title': _(
                'Edit workflow transition: %s'
            ) % self.object,
            'workflow': self.object.workflow
        }

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['workflow'] = self.object.workflow
        return kwargs

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_success_url(self):
        return reverse(
            viewname='document_states:workflow_template_transition_list',
            kwargs={'workflow_template_id': self.object.workflow.pk}
        )


class WorkflowTemplateTransitionListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = Workflow
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_template_id'
    object_permission = permission_workflow_template_view
    view_icon = icon_workflow_template_transition_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_template_transition,
            'no_results_main_link': link_workflow_template_transition_create.resolve(
                context=RequestContext(
                    dict_={'workflow': self.external_object},
                    request=self.request
                )
            ),
            'no_results_text': _(
                'Create a transition and use it to move a workflow from '
                ' one state to another.'
            ),
            'no_results_title': _(
                'This workflow doesn\'t have any transitions'
            ),
            'object': self.external_object,
            'title': _(
                'Transitions of workflow: %s'
            ) % self.external_object,
            'workflow': self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.transitions.all()


class WorkflowTemplateTransitionFieldCreateView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_class = WorkflowTransition
    external_object_permission = permission_workflow_template_edit
    external_object_pk_url_kwarg = 'workflow_template_transition_id'
    fields = (
        'name', 'label', 'field_type', 'help_text', 'required', 'widget',
        'widget_kwargs'
    )
    view_icon = icon_workflow_template_transition_field_create

    def get_extra_context(self):
        return {
            'navigation_object_list': ('transition', 'workflow'),
            'transition': self.external_object,
            'title': _(
                'Create a field for workflow transition: %s'
            ) % self.external_object,
            'workflow': self.external_object.workflow
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'transition': self.external_object
        }

    def get_queryset(self):
        return self.external_object.fields.all()

    def get_post_action_redirect(self):
        return reverse(
            viewname='document_states:workflow_template_transition_field_list',
            kwargs={
                'workflow_template_transition_id': self.external_object.pk
            }
        )


class WorkflowTemplateTransitionFieldDeleteView(SingleObjectDeleteView):
    model = WorkflowTransitionField
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_transition_field_id'
    view_icon = icon_workflow_template_transition_field_delete

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_template_transition', 'workflow'
            ),
            'object': self.object,
            'title': _('Delete workflow transition field: %s') % self.object,
            'workflow': self.object.transition.workflow,
            'workflow_template_transition': self.object.transition
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='document_states:workflow_template_transition_field_list',
            kwargs={
                'workflow_template_transition_id': self.object.transition.pk
            }
        )


class WorkflowTemplateTransitionFieldEditView(SingleObjectEditView):
    fields = (
        'name', 'label', 'field_type', 'help_text', 'required', 'widget',
        'widget_kwargs'
    )
    model = WorkflowTransitionField
    object_permission = permission_workflow_template_edit
    pk_url_kwarg = 'workflow_template_transition_field_id'
    view_icon = icon_workflow_template_transition_field_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_template_transition', 'workflow'
            ),
            'object': self.object,
            'title': _('Edit workflow transition field: %s') % self.object,
            'workflow': self.object.transition.workflow,
            'workflow_template_transition': self.object.transition
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='document_states:workflow_template_transition_field_list',
            kwargs={
                'workflow_template_transition_id': self.object.transition.pk
            }
        )


class WorkflowTemplateTransitionFieldListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_class = WorkflowTransition
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_template_transition_id'
    view_icon = icon_workflow_template_transition_field_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'navigation_object_list': ('object', 'workflow'),
            'no_results_icon': icon_workflow_template_transition_field,
            'no_results_main_link': link_workflow_template_transition_field_create.resolve(
                context=RequestContext(
                    dict_={
                        'object': self.external_object
                    }, request=self.request
                )
            ),
            'no_results_text': _(
                'Workflow transition fields allow adding data to the '
                'workflow\'s context. This additional context data can then '
                'be used by other elements of the workflow system like the '
                'workflow state actions.'
            ),
            'no_results_title': _(
                'There are no fields for this workflow transition'
            ),
            'object': self.external_object,
            'title': _(
                'Fields for workflow transition: %s'
            ) % self.external_object,
            'workflow': self.external_object.workflow
        }

    def get_source_queryset(self):
        return self.external_object.fields.all()


class WorkflowTemplateTransitionTriggerEventListView(
    ExternalObjectViewMixin, FormView
):
    external_object_class = WorkflowTransition
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_template_transition_id'
    form_class = WorkflowTransitionTriggerEventRelationshipFormSet
    view_icon = icon_workflow_template_transition_triggers

    def dispatch(self, *args, **kwargs):
        EventType.refresh()
        return super().dispatch(*args, **kwargs)

    def form_valid(self, form):
        try:
            for instance in form:
                instance.save()
        except Exception as exception:
            messages.error(
                message=_(
                    'Error updating workflow transition trigger events; %s'
                ) % exception, request=self.request

            )
        else:
            messages.success(
                message=_(
                    'Workflow transition trigger events updated successfully'
                ), request=self.request
            )

        return super().form_valid(form=form)

    def get_extra_context(self):
        return {
            'form_display_mode_table': True,
            'navigation_object_list': ('object', 'workflow'),
            'object': self.external_object,
            'subtitle': _(
                'Triggers are events that cause this transition to execute '
                'automatically.'
            ),
            'title': _(
                'Workflow transition trigger events for: %s'
            ) % self.external_object,
            'workflow': self.external_object.workflow
        }

    def get_initial(self):
        obj = self.external_object
        initial = []

        # Return the queryset by name from the sorted list of the class
        event_type_ids = [event_type.id for event_type in EventType.all()]
        event_type_queryset = StoredEventType.objects.filter(
            name__in=event_type_ids
        )

        # Sort queryset in Python by namespace, then by label
        event_type_queryset = sorted(
            event_type_queryset, key=lambda x: (x.namespace, x.label)
        )

        for event_type in event_type_queryset:
            initial.append(
                {
                    'event_type': event_type,
                    'transition': obj
                }
            )
        return initial

    def get_post_action_redirect(self):
        return reverse(
            viewname='document_states:workflow_template_transition_list',
            kwargs={'workflow_template_id': self.external_object.workflow.pk}
        )



import logging

from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig

logger = logging.getLogger(name=__name__)


class ViewsApp(MayanAppConfig):
    app_namespace = 'views'
    app_url = 'views'
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.views'
    static_media_ignore_patterns = (
        'views/node_modules/dropzone/index.js',
        'views/node_modules/dropzone/component.json'
    )
    verbose_name = _('Views')

    def ready(self):
        super().ready()



from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..models import Theme

from .literals import TEST_THEME_LABEL, TEST_THEME_LABEL_EDITED


class ThemeTestMixin(TestMixinObjectCreationTrack):
    _test_object_model = Theme
    _test_object_name = '_test_theme'

    def _create_test_theme(self):
        self._test_theme = Theme.objects.create(
            label=TEST_THEME_LABEL
        )

    def _edit_test_theme(self):
        self._test_theme.label = TEST_THEME_LABEL_EDITED
        self._test_theme.save()


class ThemeViewTestMixin(ThemeTestMixin):
    def _request_test_theme_create_view(self):
        self._test_object_track()

        response = self.post(
            viewname='appearance:theme_create', data={
                'label': TEST_THEME_LABEL,
            }
        )

        self._test_object_set()

        return response

    def _request_test_theme_delete_view(self):
        return self.post(
            viewname='appearance:theme_delete', kwargs={
                'theme_id': self._test_theme.pk
            }
        )

    def _request_test_theme_edit_view(self):
        return self.post(
            viewname='appearance:theme_edit', kwargs={
                'theme_id': self._test_theme.pk
            }, data={
                'label': TEST_THEME_LABEL_EDITED,
            }
        )

    def _request_test_theme_list_view(self):
        return self.get(viewname='appearance:theme_list')


class UserThemeSettingsViewTestMixin(ThemeTestMixin):
    def _request_test_current_user_theme_settings_detail_view(self):
        return self._request_test_user_theme_settings_detail_view(
            user=self._test_case_user
        )

    def _request_test_current_user_theme_settings_edit_view(self):
        return self._request_test_user_theme_settings_edit_view(
            user=self._test_case_user
        )

    def _request_test_super_user_theme_settings_detail_view(self):
        return self._request_test_user_theme_settings_detail_view(
            user=self._test_super_user
        )

    def _request_test_super_user_theme_settings_edit_view(self):
        return self._request_test_user_theme_settings_edit_view(
            user=self._test_super_user
        )

    def _request_test_user_theme_settings_detail_view(self, user=None):
        user = user or self._test_user

        return self.get(
            viewname='appearance:user_theme_settings_detail', kwargs={
                'user_id': user.pk
            }
        )

    def _request_test_user_theme_settings_edit_view(self, user=None):
        user = user or self._test_user

        return self.post(
            viewname='appearance:user_theme_settings_edit', kwargs={
                'user_id': user.pk
            }, data={
                'theme': self._test_theme.pk
            }
        )



import django.core.files.storage
from django.db import migrations, models

import mayan.apps.document_signatures.models


class Migration(migrations.Migration):
    dependencies = [
        ('document_signatures', '0007_auto_20180403_0702'),
    ]

    operations = [
        migrations.AlterField(
            model_name='detachedsignature',
            name='signature_file',
            field=models.FileField(
                blank=True, null=True,
                storage=django.core.files.storage.FileSystemStorage(
                    location=b'/home/rosarior/development/mayan-edms/mayan/media/document_signatures'
                ),
                upload_to=mayan.apps.document_signatures.models.upload_to,
                verbose_name='Signature file'
            ),
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('metadata', '0007_auto_20150918_0800'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documentmetadata',
            options={
                'ordering': ('metadata_type',),
                'verbose_name': 'Document metadata',
                'verbose_name_plural': 'Document metadata'
            },
        ),
    ]



import base64
import logging
import random
from urllib.parse import quote_plus

from mayan.apps.sources.exceptions import SourceActionException

from ..classes import SourceStoredFile

from .stored_file_mixins import (
    SourceBackendMixinStoredFileDeleteInteractive,
    SourceBackendMixinStoredFileDeleteInteractiveNot,
    SourceBackendMixinStoredFileDocumentFileUpload,
    SourceBackendMixinStoredFileDocumentUpload,
    SourceBackendMixinStoredFileImage, SourceBackendMixinStoredFileList
)

logger = logging.getLogger(name=__name__)


class SourceBackendMixinStoredFileSourceBase:
    @classmethod
    def intialize(cls):
        SourceStoredFile.initialize()

    def action_file_get(self, **kwargs):
        source_stored_file = self.get_stored_file(**kwargs)
        yield from (
            {'file': source_stored_file},
        )

    def get_stored_file(self, encoded_filename=None, filename=None):
        identifier = None

        if encoded_filename:
            identifier = encoded_filename
            identifier_name = 'encoded_filename'
        elif filename:
            identifier = filename
            identifier_name = 'filename'

        if not identifier:
            raise SourceActionException(
                'Must provide either `encoded_filename` or `filename`.'
            )

        for stored_file in self.get_stored_file_list():
            if getattr(stored_file, identifier_name) == identifier:
                return stored_file

        raise SourceActionException('Requested file not found.')


class SourceBackendMixinStoredFileInteractive(
    SourceBackendMixinStoredFileDeleteInteractive,
    SourceBackendMixinStoredFileDocumentFileUpload,
    SourceBackendMixinStoredFileDocumentUpload,
    SourceBackendMixinStoredFileImage,
    SourceBackendMixinStoredFileList,
    SourceBackendMixinStoredFileSourceBase
):
    """
    Class for sources with stored files that support uploading documents and
    document files.
    """


class SourceBackendMixinStoredFileInteractiveNot(
    SourceBackendMixinStoredFileDeleteInteractiveNot,
    SourceBackendMixinStoredFileSourceBase
):
    def get_file_identifier(self):
        file_list_generator = self.get_stored_file_list()

        file_list_generator = list(file_list_generator)

        if file_list_generator:
            choice = random.choice(seq=file_list_generator)

            base64_filename = base64.urlsafe_b64encode(
                s=str(choice).encode('utf8')
            )

            return quote_plus(string=base64_filename)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c

queue_duplicates = CeleryQueue(
    label=_('Duplicates'), name='duplicates', worker=worker_c
)
queue_duplicates_slow = CeleryQueue(
    label=_('Duplicates slow'), name='duplicates_slow', worker=worker_c
)

queue_duplicates.add_task_type(
    dotted_path='mayan.apps.duplicates.tasks.task_duplicates_clean_empty_lists',
    label=_('Clean empty duplicate lists')
)
queue_duplicates.add_task_type(
    dotted_path='mayan.apps.duplicates.tasks.task_duplicates_scan_for',
    label=_('Scan document duplicates')
)

queue_duplicates_slow.add_task_type(
    dotted_path='mayan.apps.duplicates.tasks.task_duplicates_scan_all',
    label=_('Duplicated document scan')
)



from django.db import migrations


def code_make_labels_unique(apps, schema_editor):
    WebLink = apps.get_model(app_label='web_links', model_name='WebLink')

    for web_link in WebLink.objects.using(alias=schema_editor.connection.alias).all():
        # Look for instances with the same label
        duplicate_queryset = WebLink.objects.using(
            alias=schema_editor.connection.alias
        ).filter(label=web_link.label).exclude(pk=web_link.pk)
        if duplicate_queryset:
            # If a duplicate is found, append the id to the original instance
            # label
            web_link.label = '{}__{}'.format(web_link.label, web_link.pk)
            web_link.save()


def code_make_labels_unique_reverse(apps, schema_editor):
    WebLink = apps.get_model(app_label='web_links', model_name='WebLink')

    for web_link in WebLink.objects.using(alias=schema_editor.connection.alias).all():
        if web_link.label.endswith('__{}'.format(web_link.pk)):
            web_link.label = web_link.label.replace(
                '__{}'.format(web_link.pk), ''
            )
            web_link.save()


class Migration(migrations.Migration):
    dependencies = [
        ('web_links', '0003_auto_20191211_0233')
    ]

    operations = [
        migrations.RunPython(
            code=code_make_labels_unique,
            reverse_code=code_make_labels_unique_reverse
        )
    ]
